# standard libraries
import contextlib
import copy
import math
import typing
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift import Application
from nion.swift import Facade
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.swift.model import DynamicString
from nion.swift.model import ImportExportManager
from nion.swift.model import Graphics
from nion.swift.test import TestContext
from nion.ui import TestUI


Facade.initialize()


def create_memory_profile_context() -> TestContext.MemoryProfileContext:
    return TestContext.MemoryProfileContext()


class TestDisplayItemClass(unittest.TestCase):

    def setUp(self):
        TestContext.begin_leaks()
        self._test_setup = TestContext.TestSetup()

    def tearDown(self):
        self._test_setup = typing.cast(typing.Any, None)
        TestContext.end_leaks(self)

    def test_display_item_with_multiple_display_data_channels_has_sensible_properties(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2, False)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.append_display_data_channel(DisplayItem.DisplayDataChannel(data_item=data_item2))
            self.assertIsNotNone(display_item.size_and_data_format_as_string)
            self.assertIsNotNone(display_item.date_for_sorting)
            self.assertIsNotNone(display_item.date_for_sorting_local_as_string)
            self.assertIsNotNone(display_item.status_str)
            self.assertIsNotNone(display_item.project_str)
            self.assertIsNotNone(display_item.used_display_type)

    def test_display_data_channel_disconnects_if_display_item_closed(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item_snapshot = display_item.snapshot()
            self.assertEqual(2, len(data_item.display_data_channels))
            display_item_snapshot.close()
            self.assertEqual(1, len(data_item.display_data_channels))

    def test_display_item_snapshot_and_copy_preserve_display_type(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_type = "line_plot"
            with contextlib.closing(display_item.snapshot()) as snapshot_display_item:
                self.assertEqual("line_plot", snapshot_display_item.display_type)
                with contextlib.closing(copy.deepcopy(display_item)) as copy_display_item:
                    self.assertEqual("line_plot", copy_display_item.display_type)

    def test_display_item_snapshot_and_preserves_display_data_channel_properties(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones((2, 3, 3, 4, 4)), data_descriptor=DataAndMetadata.DataDescriptor(True, 2, 2)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            # note: not testing all properties, only indexes
            display_item.display_data_channels[0].sequence_index = 1
            display_item.display_data_channels[0].collection_index = (1, 2)
            with contextlib.closing(display_item.snapshot()) as snapshot_display_item:
                self.assertEqual(1, snapshot_display_item.display_data_channels[0].sequence_index)
                self.assertEqual((1, 2, 0), snapshot_display_item.display_data_channels[0].collection_index)

    def test_appending_display_data_channel_does_nothing_if_display_data_channel_already_exists(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            self.assertEqual(1, len(display_item.display_data_channels))
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(display_item.display_data_channels))
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(display_item.display_data_channels))

    def test_appending_display_data_channel_adds_layer(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            self.assertEqual(1, len(display_item.display_layers))
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(display_item.display_data_channels))

    def test_appending_then_removing_display_data_channel_returns_to_original(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            self.assertEqual(1, len(display_item.display_data_channels))
            self.assertEqual(1, len(display_item.display_layers))
            self.assertEqual(data_item1, display_item.display_data_channel.data_item)
            display_data_channel_uuid = display_item.display_data_channel.uuid
            self.assertFalse(display_item.display_properties)
            self.assertEqual(1, len(display_item.display_layers))
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(display_item.display_data_channels))
            display_item.remove_display_data_channel(display_item.display_data_channels[-1]).close()
            self.assertEqual(1, len(display_item.display_data_channels))
            self.assertEqual(1, len(display_item.display_layers))
            self.assertEqual(data_item1, display_item.display_data_channel.data_item)
            self.assertEqual(display_data_channel_uuid, display_item.display_data_channel.uuid)
            self.assertFalse(display_item.display_properties)

    def test_removing_data_item_updates_display_layer_data_indexes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            display_item.append_display_data_channel_for_data_item(data_item2)
            display_item._set_display_layer_property(0, "ref", "A")
            display_item._set_display_layer_property(1, "ref", "B")
            self.assertEqual(2, len(display_item.display_data_channels))
            self.assertEqual(display_item.display_data_channels[0], display_item.get_display_layer_display_data_channel(0))
            self.assertEqual(display_item.display_data_channels[1], display_item.get_display_layer_display_data_channel(1))
            document_model.remove_data_item(data_item1)
            self.assertEqual(1, len(display_item.display_layers))
            self.assertEqual(display_item.display_data_channels[0], display_item.get_display_layer_display_data_channel(0))
            self.assertEqual("B", display_item.get_display_layer_property(0, "ref"))

    def test_removing_display_data_channel_updates_display_layer_data_indexes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item3 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            document_model.append_data_item(data_item3)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            display_item.append_display_data_channel_for_data_item(data_item2)
            display_item.append_display_data_channel_for_data_item(data_item3)
            display_item._set_display_layer_property(0, "ref", "A")
            display_item._set_display_layer_property(1, "ref", "B")
            display_item._set_display_layer_property(2, "ref", "C")
            self.assertEqual(3, len(display_item.display_data_channels))
            self.assertEqual(display_item.display_data_channels[0], display_item.get_display_layer_display_data_channel(0))
            self.assertEqual(display_item.display_data_channels[1], display_item.get_display_layer_display_data_channel(1))
            self.assertEqual(display_item.display_data_channels[2], display_item.get_display_layer_display_data_channel(2))
            display_item.remove_display_data_channel(display_item.display_data_channels[1]).close()
            self.assertEqual(2, len(display_item.display_data_channels))
            self.assertEqual(display_item.display_data_channels[0], display_item.get_display_layer_display_data_channel(0))
            self.assertEqual(display_item.display_data_channels[1], display_item.get_display_layer_display_data_channel(1))
            self.assertEqual("A", display_item.get_display_layer_property(0, "ref"))
            self.assertEqual("C", display_item.get_display_layer_property(1, "ref"))

    def test_removing_display_layer_removes_associated_display_data_channel(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(document_model.display_items))
            self.assertEqual(2, len(display_item.display_data_channels))
            self.assertEqual(2, len(display_item.display_layers))
            self.assertEqual(display_item.display_data_channels[0], display_item.display_layers[0].display_data_channel)
            self.assertEqual(display_item.display_data_channels[1], display_item.display_layers[1].display_data_channel)
            display_item.remove_display_layer(display_item.display_layers[1]).close()
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(document_model.display_items))
            self.assertEqual(1, len(display_item.display_data_channels))
            self.assertEqual(1, len(display_item.display_layers))
            self.assertEqual(display_item.display_data_channels[0], display_item.display_layers[0].display_data_channel)

    def test_removing_display_layer_removes_associated_display_data_channel_only_if_no_other_references(self):
        with TestContext.create_memory_context() as test_context:
            with TestContext.create_memory_context() as test_context:
                document_model = test_context.create_document_model()
                data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
                data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
                document_model.append_data_item(data_item1)
                document_model.append_data_item(data_item2)
                display_item = document_model.get_display_item_for_data_item(data_item1)
                display_item.append_display_data_channel_for_data_item(data_item2)
                display_item._add_display_layer_for_data_item(data_item2)
                self.assertEqual(2, len(document_model.data_items))
                self.assertEqual(2, len(document_model.display_items))
                self.assertEqual(2, len(display_item.display_data_channels))
                self.assertEqual(3, len(display_item.display_layers))
                self.assertEqual(display_item.display_data_channels[0], display_item.display_layers[0].display_data_channel)
                self.assertEqual(display_item.display_data_channels[1], display_item.display_layers[1].display_data_channel)
                self.assertEqual(display_item.display_data_channels[1], display_item.display_layers[2].display_data_channel)
                display_item.remove_display_layer(display_item.display_layers[2]).close()
                self.assertEqual(2, len(document_model.data_items))
                self.assertEqual(2, len(document_model.display_items))
                self.assertEqual(2, len(display_item.display_data_channels))
                self.assertEqual(2, len(display_item.display_layers))
                self.assertEqual(display_item.display_data_channels[0], display_item.display_layers[0].display_data_channel)
                self.assertEqual(display_item.display_data_channels[1], display_item.display_layers[1].display_data_channel)

    def test_inserting_display_data_channel_updates_display_layer_data_indexes(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            data_item3 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item3)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            display_item.append_display_data_channel_for_data_item(data_item2)
            display_item._set_display_layer_property(0, "ref", "A")
            display_item._set_display_layer_property(1, "ref", "B")
            self.assertEqual(2, len(display_item.display_data_channels))
            self.assertEqual(display_item.display_data_channels[0], display_item.get_display_layer_display_data_channel(0))
            self.assertEqual(display_item.display_data_channels[1], display_item.get_display_layer_display_data_channel(1))
            display_item.insert_display_data_channel(1, DisplayItem.DisplayDataChannel(data_item=data_item3))
            self.assertEqual(3, len(display_item.display_data_channels))
            self.assertEqual(display_item.display_data_channels[0], display_item.get_display_layer_display_data_channel(0))
            self.assertEqual(display_item.display_data_channels[2], display_item.get_display_layer_display_data_channel(1))
            self.assertEqual("A", display_item.get_display_layer_property(0, "ref"))
            self.assertEqual("B", display_item.get_display_layer_property(1, "ref"))

    def test_copy_display_item_should_copy_all_display_data_channels_and_layers(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2, False)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(1, len(document_model.display_items))
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(display_item.data_items))
            self.assertEqual(2, len(display_item.display_layers))
            display_item_copy = document_model.get_display_item_copy_new(display_item)
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(document_model.display_items))
            self.assertEqual(2, len(display_item_copy.data_items))
            self.assertEqual(2, len(display_item_copy.display_layers))
            self.assertTrue(display_item.display_layers_match(display_item_copy))

    def test_snapshot_display_item_with_data_item_and_multiple_display_layers(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((3, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_type = "line_plot"
            display_item.move_display_layer_backward(1)  # 0 1 2 -> 0 2 1
            display_item.move_display_layer_backward(0)  # 0 2 1 -> 2 0 1
            display_item.move_display_layer_backward(1)  # 2 0 1 -> 2 1 0
            self.assertEqual(1, len(document_model.display_items))
            self.assertEqual(1, len(document_model.data_items))
            self.assertEqual(1, len(display_item.data_items))
            self.assertEqual(3, len(display_item.display_layers))
            display_item_copy = document_model.get_display_item_snapshot_new(display_item)
            self.assertEqual(2, len(document_model.display_items))
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(1, len(display_item_copy.data_items))
            self.assertEqual(3, len(display_item_copy.display_layers))
            self.assertTrue(display_item.display_layers_match(display_item_copy))

    def test_add_layer_to_line_plot_with_auto_layer_color_sets_both_colors(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item)
            self.assertEqual(1, len(display_item.display_layers))
            display_item.append_display_data_channel_for_data_item(data_item2, {"stroke_color": "#000"})
            self.assertEqual(2, len(display_item.display_layers))
            self.assertIsNotNone(display_item.get_display_layer_display_data_channel(0))
            self.assertIsNotNone(display_item.get_display_layer_display_data_channel(1))
            self.assertIsNotNone(display_item.get_display_layer_property(0, "fill_color"))
            self.assertIsNotNone(display_item.get_display_layer_property(1, "stroke_color"))
            self.assertIsNone(display_item.get_display_layer_property(1, "fill_color"))

    def test_second_layer_to_line_plot_enables_caption(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item3 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2)
            document_model.append_data_item(data_item3)
            display_item = document_model.get_display_item_for_data_item(data_item)
            self.assertIsNone(display_item.get_display_property("legend_position"))
            # check that legend is automatically enabled for 2nd layer
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual("top-right", display_item.get_display_property("legend_position"))
            # check that legend is not automatically enabled for 3rd layer
            display_item.set_display_property("legend_position", None)
            display_item.append_display_data_channel_for_data_item(data_item3)
            self.assertIsNone(display_item.get_display_property("legend_position"))

    def test_closing_display_does_not_trigger_computation_binding(self):
        # this tests to make sure that items involving filters are closed properly.
        # notifications should not be sent during closing, otherwise the computation
        # will try to update its dependencies using the item being destructed.
        # this test only failed in that it printed a stack trace.
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, 8)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            fft_data_item = document_model.get_fft_new(display_item, display_item.data_item)
            document_model.recompute_all()
            fft_display_item = document_model.get_display_item_for_data_item(fft_data_item)
            document_model.get_fourier_filter_new(fft_display_item, fft_data_item)
            document_model.recompute_all()

    def test_build_table_with_calibrated_1d_data(self):
        with TestContext.create_memory_context() as test_context:
            calibration = Calibration.Calibration(1.0, 2.0, "nm")
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.array([1.1, 1.2, 1.3, 1.4]))
            data_item.set_dimensional_calibration(0, calibration)
            data_item.set_intensity_calibration(Calibration.Calibration(0, 1, "e"))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item._set_display_layer_property(0, "label", "W")
            headers, data = ImportExportManager.build_table(display_item)
            self.assertEqual(2, len(headers))
            self.assertEqual(2, len(data))
            self.assertEqual("X (nm)", headers[0])
            self.assertEqual("W (e)", headers[1])
            self.assertTrue(numpy.array_equal(calibration.convert_to_calibrated_value(numpy.arange(0, data_item.data.shape[0])), data[0]))
            self.assertTrue(numpy.array_equal(data_item.data, data[1]))

    def test_build_table_with_two_display_layers_of_same_calibrated_1d_data(self):
        with TestContext.create_memory_context() as test_context:
            calibration = Calibration.Calibration(1.0, 2.0, "nm")
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.array([1.1, 1.2, 1.3, 1.4]))
            data_item.set_dimensional_calibration(0, calibration)
            data_item.set_intensity_calibration(Calibration.Calibration(0, 1, "e"))
            data_item2 = DataItem.DataItem(numpy.array([2.1, 2.2, 2.3, 2.4]))
            data_item2.set_dimensional_calibration(0, calibration)
            data_item2.set_intensity_calibration(Calibration.Calibration(0, 1, "e"))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.append_display_data_channel_for_data_item(data_item2)
            display_item._set_display_layer_property(0, "label", "W")
            display_item._set_display_layer_property(1, "label", "T")
            headers, data = ImportExportManager.build_table(display_item)
            self.assertEqual(3, len(headers))
            self.assertEqual(3, len(data))
            self.assertEqual("X (nm)", headers[0])
            self.assertEqual("W (e)", headers[1])
            self.assertEqual("T (e)", headers[2])
            self.assertTrue(numpy.array_equal(calibration.convert_to_calibrated_value(numpy.arange(0, data_item.data.shape[0])), data[0]))
            self.assertTrue(numpy.array_equal(data_item.data, data[1]))
            self.assertTrue(numpy.array_equal(data_item2.data, data[2]))

    def test_build_table_with_two_display_layers_of_different_calibrated_1d_data(self):
        with TestContext.create_memory_context() as test_context:
            calibration = Calibration.Calibration(1.0, 2.0, "nm")
            calibration2 = Calibration.Calibration(1.5, 2.5, "nm")
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.array([1.1, 1.2, 1.3, 1.4]))
            data_item.set_dimensional_calibration(0, calibration)
            data_item.set_intensity_calibration(Calibration.Calibration(0, 1, "e"))
            data_item2 = DataItem.DataItem(numpy.array([2.1, 2.2, 2.3, 2.4]))
            data_item2.set_dimensional_calibration(0, calibration2)
            data_item2.set_intensity_calibration(Calibration.Calibration(0, 1, "e"))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.append_display_data_channel_for_data_item(data_item2)
            display_item._set_display_layer_property(0, "label", "W")
            display_item._set_display_layer_property(1, "label", "T")
            headers, data = ImportExportManager.build_table(display_item)
            self.assertEqual(4, len(headers))
            self.assertEqual(4, len(data))
            self.assertEqual("X W (nm)", headers[0])
            self.assertEqual("Y W (e)", headers[1])
            self.assertEqual("X T (nm)", headers[2])
            self.assertEqual("Y T (e)", headers[3])
            self.assertTrue(numpy.array_equal(calibration.convert_to_calibrated_value(numpy.arange(0, data_item.data.shape[0])), data[0]))
            self.assertTrue(numpy.array_equal(data_item.data, data[1]))
            self.assertTrue(numpy.array_equal(calibration2.convert_to_calibrated_value(numpy.arange(0, data_item.data.shape[0])), data[2]))
            self.assertTrue(numpy.array_equal(data_item2.data, data[3]))

    def test_display_layer_property_changed(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, )))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_layer = typing.cast(DisplayItem.DisplayLayer, display_item.display_layers[0])
            property_did_change = False
            def property_changed(name: str) -> None:
                nonlocal property_did_change
                property_did_change = True
            with contextlib.closing(display_layer.property_changed_event.listen(property_changed)):
                display_layer.fill_color = "red"
            self.assertTrue(property_did_change)

    def test_display_layer_property_reloads_after_change(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                display_layer = typing.cast(DisplayItem.DisplayLayer, display_item.display_layers[0])
                display_layer.fill_color = "red"
                display_item.title = "red"
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = document_model.data_items[0]
                display_item = document_model.get_display_item_for_data_item(data_item)
                display_layer = typing.cast(DisplayItem.DisplayLayer, display_item.display_layers[0])
                self.assertEqual("red", display_layer.fill_color)
                display_layer.fill_color = "blue"  # should be able to set value after reload. failed once.

    def test_display_layer_property_reloads_with_no_display_data_channel(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                display_layer = typing.cast(DisplayItem.DisplayLayer, display_item.display_layers[0])
                display_layer.display_data_channel = None
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = document_model.data_items[0]
                display_item = document_model.get_display_item_for_data_item(data_item)
                display_layer = typing.cast(DisplayItem.DisplayLayer, display_item.display_layers[0])
                self.assertIsNone(display_layer.display_data_channel)

    def test_display_layers_reload_after_inserting_and_removing(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                # add a new layer
                display_item.add_display_layer_for_display_data_channel(display_item.display_data_channels[0])
                self.assertEqual(2, len(display_item.display_layers))
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = document_model.data_items[0]
                display_item = document_model.get_display_item_for_data_item(data_item)
                self.assertEqual(2, len(display_item.display_layers))
                # now remove a layer
                display_item.remove_display_layer(display_item.display_layers[1]).close()
                self.assertEqual(1, len(display_item.display_layers))
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = document_model.data_items[0]
                display_item = document_model.get_display_item_for_data_item(data_item)
                self.assertEqual(1, len(display_item.display_layers))

    def test_display_layer_delta_stream_updates_after_inserting_removing_layers(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                # add a new layer
                display_item.add_display_layer_for_display_data_channel(display_item.display_data_channels[0])
                self.assertEqual(2, len(display_item.display_layers))
                # now insert/remove
                display_item.insert_display_layer_for_display_data_channel(0, display_item.display_data_channels[0])
                display_item.remove_display_layer(2).close()

                change_count = 0

                def handle_display_data_delta_stream_change(display_data_delta: DisplayItem.DisplayDataDelta) -> None:
                    nonlocal change_count
                    change_count += 1

                with display_item.display_data_delta_stream.value_stream.listen(handle_display_data_delta_stream_change):
                    display_item.display_layers[0].stroke_width = 2
                    display_item.display_layers[1].stroke_width = 2

                self.assertEqual(2, change_count)

    def test_display_layer_delta_stream_updates_after_inserting_removing_graphics(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                display_item.add_graphic(Graphics.PointGraphic())
                display_item.add_graphic(Graphics.PointGraphic())
                self.assertEqual(2, len(display_item.graphics))
                # now insert/remove
                display_item.insert_graphic(0, Graphics.PointGraphic())
                display_item.remove_graphic(display_item.graphics[-1]).close()
                self.assertEqual(2, len(display_item.graphics))

                change_count = 0

                def handle_display_data_delta_stream_change(display_data_delta: DisplayItem.DisplayDataDelta) -> None:
                    nonlocal change_count
                    change_count += 1

                with display_item.display_data_delta_stream.value_stream.listen(handle_display_data_delta_stream_change):
                    display_item.graphics[0].label = "label"
                    display_item.graphics[1].label = "label"

                self.assertEqual(2, change_count)

    def test_displayed_title_is_inherited_from_source(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                data_item.title = "red"
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_item2 = document_model.get_invert_new(display_item, display_item.data_item)
                display_item2 = document_model.get_display_item_for_data_item(data_item2)
                data_item3 = document_model.get_invert_new(display_item2, display_item2.data_item)
                display_item3 = document_model.get_display_item_for_data_item(data_item3)
                document_model.recompute_all()
                self.assertEqual("red", display_item.displayed_title)
                self.assertEqual("red (Negate)", display_item2.displayed_title)
                self.assertEqual("red (Negate) (Negate)", display_item3.displayed_title)
                data_item.title = "green"
                self.assertEqual("green", display_item.displayed_title)
                self.assertEqual("green (Negate)", display_item2.displayed_title)
                self.assertEqual("green (Negate) (Negate)", display_item3.displayed_title)

    def test_display_item_handles_invalid_calibration(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.zeros((8,)), dimensional_calibrations=[Calibration.Calibration(0.0, -math.inf, "x")]))
            document_model.append_data_item(data_item)

    def test_displayed_title_shows_dynamic_title_or_title_if_set(self):
        # requirement: dynamic_titles
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                data_item.dynamic_title = DynamicString._TestDynamicString()
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                self.assertEqual("green", display_item.displayed_title)
                data_item.title = "red"
                self.assertEqual("red", display_item.displayed_title)
                data_item.title = str()
                self.assertEqual("green", display_item.displayed_title)
                data_item.dynamic_title = DynamicString._TestDynamicString()

    def test_dynamic_title_enabled_when_setting_dynamic_title_and_disabled_when_setting_title_directly(self):
        # requirement: dynamic_titles
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                data_item.dynamic_title = DynamicString._TestDynamicString()
                self.assertTrue(data_item.dynamic_title_enabled)
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                self.assertEqual("green", display_item.displayed_title)
                data_item.title = "red"
                self.assertFalse(data_item.dynamic_title_enabled)
                self.assertEqual("red", display_item.displayed_title)
                data_item.dynamic_title = DynamicString._TestDynamicString()
                self.assertTrue(data_item.dynamic_title_enabled)

    def test_dynamic_title_updates_title_only_if_dynamic_title_is_enabled_and_updates_when_re_enabled(self):
        # requirement: dynamic_titles
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                data_item.dynamic_title = DynamicString._TestDynamicString()
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                self.assertEqual("green", display_item.displayed_title)
                data_item.dynamic_title.string_stream.value = "blue"
                self.assertEqual("blue", display_item.displayed_title)
                data_item.dynamic_title_enabled = False
                data_item.dynamic_title.string_stream.value = "purple"
                self.assertEqual("blue", display_item.displayed_title)
                data_item.dynamic_title_enabled = True
                self.assertEqual("purple", display_item.displayed_title)

    def test_displayed_title_shows_source_title_with_suffix_if_not_specified(self):
        # requirement: dynamic_titles
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                data_item.dynamic_title = DynamicString._TestDynamicString()
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_item2 = document_model.get_invert_new(display_item, display_item.data_item)
                display_item2 = document_model.get_display_item_for_data_item(data_item2)
                self.assertEqual("green (Negate)", display_item2.displayed_title)
                data_item2.title = "blue"
                self.assertEqual("blue", display_item2.displayed_title)
                data_item2.title = None
                self.assertEqual("green (Negate)", display_item2.displayed_title)
                display_item2.title = "yellow"
                self.assertEqual("yellow", display_item2.displayed_title)
                display_item2.title = None
                self.assertEqual("green (Negate)", display_item2.displayed_title)

    def test_placeholder_title_ignores_specified_data_item_or_display_item_titles(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,)))
                data_item.title = "green"
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_item2 = document_model.get_invert_new(display_item, display_item.data_item)
                display_item2 = document_model.get_display_item_for_data_item(data_item2)
                self.assertEqual("green (Negate)", display_item2.placeholder_title)
                data_item2.title = "blue"
                self.assertEqual("green (Negate)", display_item2.placeholder_title)
                data_item2.title = None
                self.assertEqual("green (Negate)", display_item2.placeholder_title)
                display_item2.title = "yellow"
                self.assertEqual("green (Negate)", display_item2.placeholder_title)
                display_item2.title = None
                self.assertEqual("green (Negate)", display_item2.placeholder_title)

    def test_calibrated_dimensional_calibrations(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8, ), numpy.uint32))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                self.assertIsNone(display_item.calibrated_dimensional_calibrations)
                data_item.dimensional_calibrations = [Calibration.Calibration(3.0, 0.5, "nm")]
                self.assertEqual(1, len(display_item.calibrated_dimensional_calibrations))
                self.assertEqual("nm", display_item.calibrated_dimensional_calibrations[0].units)
                display_item.calibration_style_id = "pixels-center"
                self.assertEqual("nm", display_item.calibrated_dimensional_calibrations[0].units)

    def test_data_info_for_calibrated_image(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((4, 4), numpy.uint32))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_info = display_item.data_info
                self.assertEqual("4, 4", data_info.data_shape_str)
                self.assertIsNone(None, data_info.calibrated_dimensional_calibrations_str)
                data_item.set_dimensional_calibration(0, Calibration.Calibration(3.0, 0.5, "nm"))
                data_info = display_item.data_info
                self.assertEqual("4, 4", data_info.data_shape_str)
                self.assertEqual("2 nm, 4", data_info.calibrated_dimensional_calibrations_str)
                data_item.set_dimensional_calibration(1, Calibration.Calibration(3.0, 0.5, "nm"))
                data_info = display_item.data_info
                self.assertEqual("4, 4", data_info.data_shape_str)
                self.assertEqual("2 nm, 2 nm", data_info.calibrated_dimensional_calibrations_str)
                data_item.set_dimensional_calibration(0, Calibration.Calibration())
                data_info = display_item.data_info
                self.assertEqual("4, 4", data_info.data_shape_str)
                self.assertEqual("4, 2 nm", data_info.calibrated_dimensional_calibrations_str)
                data_item.set_dimensional_calibration(0, Calibration.Calibration(3.0, 0.5, "nm"))
                display_item.calibration_style_id = "pixels-center"
                data_info = display_item.data_info
                self.assertEqual("4, 4", data_info.data_shape_str)
                self.assertEqual("2 nm, 2 nm", data_info.calibrated_dimensional_calibrations_str)

    def test_data_info_for_calibrated_line_plot(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_info = display_item.data_info
                self.assertEqual("8", data_info.data_shape_str)
                self.assertIsNone(None, data_info.calibrated_dimensional_calibrations_str)
                data_item.set_dimensional_calibration(0, Calibration.Calibration(3.0, 0.5, "nm"))
                data_info = display_item.data_info
                self.assertEqual("8", data_info.data_shape_str)
                self.assertEqual("4 nm", data_info.calibrated_dimensional_calibrations_str)
                display_item.calibration_style_id = "pixels-center"
                data_info = display_item.data_info
                self.assertEqual("8", data_info.data_shape_str)
                self.assertEqual("4 nm", data_info.calibrated_dimensional_calibrations_str)

    def test_data_info_for_composite_display(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                data_item = DataItem.DataItem(numpy.zeros((8, ), numpy.uint32))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_info = display_item.data_info
                self.assertEqual("8", data_info.data_shape_str)
                self.assertIsNone(data_info.calibrated_dimensional_calibrations_str)
                data_item.set_dimensional_calibration(0, Calibration.Calibration(0.0, 2.0, "nm"))
                data_info = display_item.data_info
                self.assertEqual("8", data_info.data_shape_str)
                self.assertEqual("16 nm", data_info.calibrated_dimensional_calibrations_str)
                data_item2 = DataItem.DataItem(numpy.zeros((8, ), numpy.uint32))
                document_model.append_data_item(data_item2)
                display_item.append_display_data_channel_for_data_item(data_item2)
                data_item2.set_dimensional_calibration(0, Calibration.Calibration(8.0, 4.0, "nm"))
                data_info = display_item.data_info
                self.assertEqual("20", data_info.data_shape_str)
                self.assertEqual("40 nm", data_info.calibrated_dimensional_calibrations_str)
                data_item3 = DataItem.DataItem(numpy.zeros((4, 4), numpy.uint32))
                document_model.append_data_item(data_item3)
                data_item3.set_dimensional_calibration(0, Calibration.Calibration(0.0, 1.0, "nm"))
                data_item3.set_dimensional_calibration(1, Calibration.Calibration(0.0, 1.0, "nm"))
                display_item3 = document_model.get_display_item_for_data_item(data_item3)
                data_info = display_item3.data_info
                self.assertEqual("4, 4", data_info.data_shape_str)
                self.assertEqual("4 nm, 4 nm", data_info.calibrated_dimensional_calibrations_str)

    def test_data_info_compound_data_items(self):
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                # test non-sequence of spectrum image data
                data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones((3, 3, 4)),
                                                                                         data_descriptor=DataAndMetadata.DataDescriptor(
                                                                                             False, 2, 1)))
                data_item.set_dimensional_calibration(0, Calibration.Calibration(3.0, 0.5, "nm"))
                data_item.set_dimensional_calibration(1, Calibration.Calibration(3.0, 0.5, "nm"))
                data_item.set_dimensional_calibration(2, Calibration.Calibration(3.0, 0.5, "eV"))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_info = display_item.data_info
                self.assertEqual("3, 3", data_info.data_shape_str)
                self.assertEqual("1.5 nm, 1.5 nm", data_info.calibrated_dimensional_calibrations_str)
                # test non-sequence of compound data
                data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones((3, 3, 4, 4)),
                                                                                         data_descriptor=DataAndMetadata.DataDescriptor(
                                                                                             False, 2, 2)))
                data_item.set_dimensional_calibration(0, Calibration.Calibration(3.0, 0.5, "nm"))
                data_item.set_dimensional_calibration(1, Calibration.Calibration(3.0, 0.5, "nm"))
                data_item.set_dimensional_calibration(2, Calibration.Calibration(3.0, 0.5, "rad"))
                data_item.set_dimensional_calibration(3, Calibration.Calibration(3.0, 0.5, "rad"))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_info = display_item.data_info
                self.assertEqual("4, 4", data_info.data_shape_str)
                self.assertEqual("2 rad, 2 rad", data_info.calibrated_dimensional_calibrations_str)
                # test sequence of compound data
                data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones((2, 3, 3, 4, 4)),
                                                                                         data_descriptor=DataAndMetadata.DataDescriptor(
                                                                                             True, 2, 2)))
                data_item.set_dimensional_calibration(0, Calibration.Calibration(3.0, 0.5, "s"))
                data_item.set_dimensional_calibration(1, Calibration.Calibration(3.0, 0.5, "nm"))
                data_item.set_dimensional_calibration(2, Calibration.Calibration(3.0, 0.5, "nm"))
                data_item.set_dimensional_calibration(3, Calibration.Calibration(3.0, 0.5, "rad"))
                data_item.set_dimensional_calibration(4, Calibration.Calibration(3.0, 0.5, "rad"))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                data_info = display_item.data_info
                self.assertEqual("4, 4", data_info.data_shape_str)
                self.assertEqual("2 rad, 2 rad", data_info.calibrated_dimensional_calibrations_str)

    def test_display_data_shape_calculator(self) -> None:
        l = [
            ((0, 1), (3, 2), False, 0, 2),          # image
            ((1, 2), (3, 2, 4), True, 0, 2),        # sequence of image
            ((1, 2), (3, 2, 4), False, 1, 2),       # 1d collection of image
            ((2, 3), (5, 3, 2, 4), True, 1, 2),     # sequence of 1d collection of image
            ((0, 1), (3, 2, 4), False, 2, 1),       # 2d collection of spectra (SI)
            ((1, 2), (5, 3, 2, 4), True, 2, 1),     # sequence of collection of spectra (SI)
            ((2, 3), (2, 3, 4, 5), False, 2, 2),    # 2d collection of image
            ((3, 4), (2, 3, 4, 5, 6), True, 2, 2),  # sequence of 2d collection of image
            ((0,), (3,), False, 0, 1),              # spectra
            ((1,), (2, 3), True, 0, 1),             # sequence of spectra
            ((1,), (2, 3), False, 1, 1),            # 1d collection of spectra
            ((2,), (2, 3, 4), True, 1, 1),          # sequence of 1d collection of spectra
        ]
        with create_memory_profile_context() as profile_context:
            document_model = profile_context.create_document_model(auto_close=False)
            with document_model.ref():
                for expected, shape, is_sequence, collection_dimension_count, datum_dimension_count in l:
                    data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones(shape), data_descriptor=DataAndMetadata.DataDescriptor(is_sequence, collection_dimension_count, datum_dimension_count)))
                    document_model.append_data_item(data_item)
                    self.assertEqual(len(expected), len(document_model.display_items[-1].displayed_display_data_calibrations))
                    self.assertEqual(expected, DisplayItem.DisplayDataShapeCalculator(data_item.data_metadata).indexes)

    # test_transaction_does_not_cascade_to_data_item_refs
    # test_increment_data_ref_counts_cascades_to_data_item_refs
    # test_adding_data_item_twice_to_composite_item_fails
    # test_composition_item_starts_drag_with_composition_item_mime_data
    # test_composite_library_item_produces_composite_display
    # test_changing_display_type_of_composite_updates_displays_in_canvas_item
    # test_changing_display_type_of_child_updates_composite_display
    # test_composite_item_deletes_cleanly_when_displayed
    # test_delete_composite_cascade_delete_works
    # test_creating_r_var_on_composite_item
    # test_creating_r_var_on_library_items
    # test_transaction_on_composite_display_propagates_to_dependents
    # test_composite_item_deletes_children_when_deleted
    # test_undelete_composite_item
    # test_composite_line_plot_initializes_properly
    # test_composite_line_plot_calculates_calibrated_data_of_two_data_items_with_same_units_but_different_scales_properly
    # test_composite_line_plot_handles_drawing_with_fixed_y_scale_and_without_data
    # test_composite_line_plot_handles_first_components_without_data
    # test_multi_line_plot_without_calibration_does_not_display_any_line_graphs
    # test_multi_line_plot_handles_calibrated_vs_uncalibrated_display
    # test_delete_and_undelete_from_memory_storage_system_restores_composite_item_after_reload
    # test_data_item_with_references_to_another_data_item_reloads
    # test_composite_library_item_reloads_metadata
    # test_composite_data_item_saves_to_file_storage
    # test_composition_display_thumbnail_source_produces_library_item_mime_data


if __name__ == '__main__':
    unittest.main()



"""
See color maps:
    https://sciviscolor.org/
    http://www.kennethmoreland.com/color-advice/
    https://datascience.lanl.gov/colormaps.html
"""

import colorsys
import dataclasses
import gettext
import json
import pathlib

import numpy
import numpy.typing
import math
import os
import pkgutil
import re
import typing
import xml.etree.ElementTree

from nion.utils import Registry

_ = gettext.gettext

_LookupDataArray = numpy.typing.NDArray[typing.Any]
_RGBA8ImageDataType = numpy.typing.NDArray[typing.Any]
_PointsType = typing.List[typing.Dict[str, typing.Union[float, typing.Tuple[int, int, int]]]]


def interpolate_colors(array: _LookupDataArray, x: int) -> _RGBA8ImageDataType:
    """
    Creates a color map for values in array
    :param array: color map to interpolate
    :param x: number of colors
    :return: interpolated color map
    """
    out_array: typing.List[typing.Tuple[int, int, int]] = []
    for i in range(x):
        if i % (x / (len(array) - 1)) == 0:
            index = i / (x / (len(array) - 1))
            out_array.append(array[int(index)])
        else:
            start_marker = array[math.floor(i / (x / (len(array) - 1)))]
            stop_marker = array[math.ceil(i / (x / (len(array) - 1)))]
            interp_amount = i % (x / (len(array) - 1)) / (x / (len(array) - 1))
            interp_color = numpy.rint(start_marker + ((stop_marker - start_marker) * interp_amount))
            out_array.append(interp_color)
    out_array[-1] = array[-1]
    return numpy.array(out_array).astype(numpy.uint8)


def generate_lookup_array_from_points(points: _PointsType, n: int) -> _RGBA8ImageDataType:
    assert points[0]["x"] == 0.0
    assert points[-1]["x"] == 1.0
    out_array: typing.List[numpy.typing.NDArray[typing.Any]] = []
    last_ix = None
    last_rgb = None
    for point in points:
        if "rgb" in point:
            r, g, b = typing.cast(typing.Tuple[int, int, int], point["rgb"])
        else:
            r = round(typing.cast(float, point["r"]) * 255)
            g = round(typing.cast(float, point["g"]) * 255)
            b = round(typing.cast(float, point["b"]) * 255)
        x = typing.cast(float, point["x"])
        assert 0 <= r <= 255
        assert 0 <= g <= 255
        assert 0 <= b <= 255
        assert 0 <= x <= 1
        rgb: numpy.typing.NDArray[typing.Any] = numpy.array([b, g, r])
        ix = int(math.floor(x * (n - 1)))
        if last_ix is None:
            out_array.append(numpy.copy(rgb))
        elif ix > last_ix:
            amount = (rgb - last_rgb) / (ix - last_ix)
            for x in range(ix - last_ix):
                out_array.append(numpy.rint(last_rgb + amount * (x + 1)))
        else:
            assert ix >= last_ix
        last_ix = ix
        last_rgb = numpy.copy(rgb)
    return numpy.array(out_array).astype(numpy.uint8)


def generate_lookup_array_grayscale() -> _RGBA8ImageDataType:
    out_list = []
    for i in range(256):
        out_list.append([i, i, i])
    return numpy.array(out_list).astype(numpy.uint8)

lookup_arrays = {
    'magma':     [[0, 0, 0],
                  [127, 0, 127],
                  [96, 96, 255],
                  [64, 192, 255],
                  [223, 255, 255]],
    'viridis':   [[32, 32, 64],
                  [160, 160, 64],
                  [208, 208, 0],
                  [144, 255, 32],
                  [0, 255, 255]],
    'plasma':    [[127, 0, 0],
                  [127, 0, 127],
                  [127, 127, 255],
                  [127, 255, 255],
                  [195, 255, 255]],
    'ice':       [[0, 0, 0],
                  [127, 0, 0],
                  [255, 127, 0],
                  [255, 255, 127],
                  [255, 255, 255]]
}

def generate_lookup_array(color_map_id: str) -> _RGBA8ImageDataType:
    return interpolate_colors(numpy.array(lookup_arrays[color_map_id]), 256)

def generate_lookup_array_hsv() -> _RGBA8ImageDataType:
    result_array = []
    for lookup_value in range(256):
        color_values = [x * 255 for x in colorsys.hsv_to_rgb(lookup_value / 300, 1.0, 1.0)]
        result_array.append(color_values)
    return numpy.array(result_array).astype(numpy.uint8)


@dataclasses.dataclass
class ColorMap:
    color_map_id: str
    name: str
    data: _RGBA8ImageDataType


color_maps: typing.Dict[str, ColorMap] = dict()

def add_color_map(color_map: ColorMap) -> None:
    color_maps[color_map.color_map_id] = color_map

add_color_map(ColorMap("grayscale", _("Grayscale"), generate_lookup_array_grayscale()))
add_color_map(ColorMap("magma", _("Magma"), generate_lookup_array('magma')))
add_color_map(ColorMap("hsv", _("HSV"), generate_lookup_array_hsv()))
add_color_map(ColorMap("viridis", _("Viridis"), generate_lookup_array('viridis')))
add_color_map(ColorMap("plasma", _("Plasma"), generate_lookup_array('plasma')))
add_color_map(ColorMap("ice", _("Ice"), generate_lookup_array('ice')))


def get_color_map_id_from_name(name: str) -> str:
    color_map_id = name.lower()
    color_map_id = re.sub(r"[^\w\s]", '', color_map_id)
    color_map_id = re.sub(r"\s+", '-', color_map_id)
    return color_map_id


def load_color_map_json_str(color_map_json_str: str) -> None:
    color_map_json = json.loads(color_map_json_str)
    color_map_id = color_map_json["id"]
    add_color_map(ColorMap(color_map_id, color_map_json["name"], generate_lookup_array_from_points(color_map_json["points"], 256)))


def load_color_map_xml_str(color_map_xml_str: str, name: str, color_map_id: str) -> None:
    tree_root = xml.etree.ElementTree.fromstring(color_map_xml_str)
    assert tree_root.tag == "ColorMaps"
    color_map_tree = list(tree_root)[0]
    assert color_map_tree.tag == "ColorMap"
    raw_points = [point_tree.attrib for point_tree in color_map_tree]
    points: _PointsType = list()
    for raw_point in raw_points:
        if "x" in raw_point:
            points.append({"x": float(raw_point['x']), "r": float(raw_point['r']), "g": float(raw_point['g']),
                           "b": float(raw_point['b'])})
    add_color_map(ColorMap(color_map_id, name, generate_lookup_array_from_points(points, 256)))
    """
    # this section can be used to generate .json from .xml color tables
    points2 = [{"x": point['x'], "rgb": [round(point['r'] * 255), round(point['g'] * 255), round(point['b'] * 255)]} for point in points]
    s = ""
    s += '{' + '\n'
    s += f'  "id": "{color_map_id}",' + '\n'
    s += f'  "name": "{name}",' + '\n'
    s += '  "points": [' + '\n'
    bro = '{'
    brc = '}'
    for point2 in points2[:-1]:
        s += f'    {bro}"x": {point2["x"]}, "rgb": [{point2["rgb"][0]}, {point2["rgb"][1]}, {point2["rgb"][2]}]{brc},' + '\n'
    point2 = points2[-1]
    s += f'    {bro}"x": {point2["x"]}, "rgb": [{point2["rgb"][0]}, {point2["rgb"][1]}, {point2["rgb"][2]}]{brc}' + '\n'
    s += '  ]' + '\n'
    s += '}' + '\n'
    print(s)
    # d = {"id": color_map_id, "name": name, "points": points2}
    # print(json.dumps(d, indent=2))
    """


def load_color_maps(color_maps_dir: pathlib.Path) -> None:
    for root, dirs, files in os.walk(color_maps_dir):
        for file in files:
            color_map_path = pathlib.Path(root) / file
            if not color_map_path.name.startswith("."):
                color_map_file_extension = color_map_path.suffix
                try:
                    if color_map_file_extension == ".json":
                        with open(color_map_path, "r") as f:
                            load_color_map_json_str(f.read())
                    elif color_map_file_extension == ".xml":
                        with open(color_map_path, "r") as f:
                            xml_str = f.read()
                            load_color_map_xml_str(xml_str, color_map_path.stem, get_color_map_id_from_name(color_map_path.stem))
                except Exception as e:
                    import traceback
                    traceback.print_exc()


def load_color_map_from_dict(d: typing.Mapping[str, typing.Any]) -> None:
    color_map_id = d["id"]
    color_map_name = d["name"]
    color_map_points = d["points"]
    add_color_map(ColorMap(color_map_id, color_map_name, generate_lookup_array_from_points(color_map_points, 256)))


def load_color_map_resource(resource_path: str) -> None:
    bytes = pkgutil.get_data(__name__, resource_path)
    assert bytes is not None
    load_color_map_json_str(bytes.decode("utf-8"))


load_color_map_resource("resources/color_maps/black_body.json")
load_color_map_resource("resources/color_maps/extended_black_body.json")
load_color_map_resource("resources/color_maps/extended_kindlmann.json")
load_color_map_resource("resources/color_maps/kindlmann.json")


def get_color_map_data_by_id(color_map_id: str) -> _RGBA8ImageDataType:
    return color_maps.get(color_map_id, color_maps["grayscale"]).data


class ColorMapProtocol(typing.Protocol):
    color_map_id: str
    name: str
    data: typing.Optional[_RGBA8ImageDataType] = None
    json_str: typing.Optional[str] = None
    xml_str: typing.Optional[str] = None


def component_registered(component: Registry._ComponentType, component_types: typing.Set[str]) -> None:
    if "color-map-description" in component_types:
        color_map_description = typing.cast(ColorMapProtocol, component)
        if hasattr(color_map_description, "data") and color_map_description.data is not None:
            add_color_map(ColorMap(color_map_description.color_map_id, color_map_description.name, color_map_description.data))
        elif hasattr(color_map_description, "json_str") and color_map_description.json_str is not None:
            load_color_map_json_str(color_map_description.json_str)
        elif hasattr(color_map_description, "xml_str") and color_map_description.xml_str is not None:
            load_color_map_xml_str(color_map_description.xml_str, color_map_description.name, color_map_description.color_map_id)


def component_unregistered(component: Registry._ComponentType, component_types: typing.Set[str]) -> None:
    if "color-map-description" in component_types:
        pass


_component_registered_listener = Registry.listen_component_registered_event(component_registered)
_component_unregistered_listener = Registry.listen_component_unregistered_event(component_unregistered)

Registry.fire_existing_component_registered_events("color-map-dict")
Registry.fire_existing_component_registered_events("color-map-json-str")
Registry.fire_existing_component_registered_events("color-map-xml-str")



# standard libraries
import contextlib
import typing
import unittest

# third party libraries
import numpy

# local libraries
from nion.swift import Application
from nion.swift import ComputationPanel
from nion.swift import Facade
from nion.swift import Inspector
from nion.swift.model import DataItem
from nion.swift.model import Graphics
from nion.swift.model import Symbolic
from nion.swift.test import TestContext
from nion.ui import TestUI


Facade.initialize()


def create_memory_profile_context() -> TestContext.MemoryProfileContext:
    return TestContext.MemoryProfileContext()


class TestComputationPanelClass(unittest.TestCase):

    def setUp(self):
        TestContext.begin_leaks()
        self._test_setup = TestContext.TestSetup()

    def tearDown(self):
        self._test_setup = typing.cast(typing.Any, None)
        TestContext.end_leaks(self)

    def test_expression_updates_when_node_is_changed(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item1 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation("target.xdata = -a.xdata")
            computation.create_input_item("a", Symbolic.make_item(data_item1))
            document_model.set_data_item_computation(data_item2, computation)
            panel = ComputationPanel.EditComputationDialog(document_controller, data_item2)
            document_controller.periodic()  # execute queue
            text1 = panel._text_edit_for_testing.text
            document_model.get_data_item_computation(data_item2).expression = "target.xdata = -a.xdata + 1"
            document_controller.periodic()  # execute queue
            text2 = panel._text_edit_for_testing.text
            self.assertNotEqual(text2, text1)

    def test_clearing_computation_clears_text_and_unbinds_or_whatever(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item1 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation("target.xdata = -a.xdata")
            computation.create_input_item("a", Symbolic.make_item(data_item1))
            document_model.set_data_item_computation(data_item2, computation)
            panel = ComputationPanel.EditComputationDialog(document_controller, data_item2)
            document_controller.periodic()  # execute queue
            panel._text_edit_for_testing.text = ""  # no longer clears the computation. cm 2020-08.
            panel._update_button.on_clicked()
            document_controller.periodic()
            self.assertIsNotNone(document_model.get_data_item_computation(data_item2))
            text2 = panel._text_edit_for_testing.text
            self.assertFalse(text2)

    def test_invalid_expression_shows_error_and_clears_it(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item1 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation("target.xdata = -a.xdata")
            computation.create_input_item("a", Symbolic.make_item(data_item1))
            document_model.set_data_item_computation(data_item2, computation)
            panel = ComputationPanel.EditComputationDialog(document_controller, data_item2)
            document_controller.periodic()  # let the inspector see the computation
            document_controller.periodic()  # and update the computation
            expression = panel._text_edit_for_testing.text
            self.assertFalse(panel._error_label_for_testing.text.strip())
            panel._text_edit_for_testing.text = "target.xdata = xyz(a.xdata)"
            panel._update_button.on_clicked()
            # the sequence of periodic/recompute_all is intentional, to test various computation states
            document_controller.periodic()
            document_model.recompute_all()
            document_model.recompute_all()
            document_controller.periodic()
            self.assertEqual(panel._text_edit_for_testing.text, "target.xdata = xyz(a.xdata)")
            self.assertTrue(len(panel._error_label_for_testing.text) > 0)
            panel._text_edit_for_testing.text = expression
            panel._update_button.on_clicked()
            # the sequence of periodic/recompute_all is intentional, to test various computation states
            document_controller.periodic()
            document_model.recompute_all()
            document_model.recompute_all()
            document_controller.periodic()
            self.assertEqual(panel._text_edit_for_testing.text, expression)
            self.assertIsNone(panel._error_label_for_testing.text)

    def test_error_text_cleared_after_invalid_script_becomes_valid(self):
        # similar to test_invalid_expression_shows_error_and_clears_it except periodic occurs before recompute at end
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item1 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation("target.xdata = -a.xdata")
            computation.create_input_item("a", Symbolic.make_item(data_item1))
            document_model.set_data_item_computation(data_item2, computation)
            panel = ComputationPanel.EditComputationDialog(document_controller, data_item2)
            document_controller.periodic()  # let the inspector see the computation
            document_controller.periodic()  # and update the computation
            expression = panel._text_edit_for_testing.text
            self.assertFalse(panel._error_label_for_testing.text.strip())
            panel._text_edit_for_testing.text = "target.xdata = xyz(a.xdata)"
            panel._update_button.on_clicked()
            # the sequence of periodic/recompute_all is intentional, to test various computation states
            document_controller.periodic()
            document_model.recompute_all()
            document_model.recompute_all()
            document_controller.periodic()
            self.assertEqual(panel._text_edit_for_testing.text, "target.xdata = xyz(a.xdata)")
            self.assertTrue(len(panel._error_label_for_testing.text) > 0)
            panel._text_edit_for_testing.text = expression
            panel._update_button.on_clicked()
            # the sequence of periodic/recompute_all is intentional, to test various computation states
            document_controller.periodic()
            document_model.recompute_all()
            document_model.recompute_all()
            document_controller.periodic()
            self.assertEqual(panel._text_edit_for_testing.text, expression)
            self.assertIsNone(panel._error_label_for_testing.text)

    def test_variables_get_updates_when_switching_data_items(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item1 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation("target.xdata = a.xdata + x")
            computation.create_input_item("a", Symbolic.make_item(data_item1))
            computation.create_variable("x", value_type="integral", value=5)
            document_model.set_data_item_computation(data_item2, computation)
            panel1 = ComputationPanel.EditComputationDialog(document_controller, data_item1)
            with contextlib.closing(panel1):
                document_controller.periodic()  # execute queue
                self.assertEqual(len(panel1._sections_for_testing), 0)
                panel2 = ComputationPanel.EditComputationDialog(document_controller, data_item2)
                with contextlib.closing(panel2):
                    document_controller.periodic()  # execute queue
                    self.assertEqual(len(panel2._sections_for_testing), 2)
                    document_controller.periodic()  # execute queue
                    self.assertEqual(len(panel1._sections_for_testing), 0)

    def test_change_variable_command_resulting_in_error_undo_redo(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            # setup
            data_item1 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item2)
            data_item3 = DataItem.DataItem(numpy.zeros((5, )))
            document_model.append_data_item(data_item3)
            computation = document_model.create_computation("target.xdata = a.xdata[6] + x")
            variable = computation.create_input_item("a", Symbolic.make_item(data_item2))
            computation.create_variable("x", value_type="integral", value=5)
            document_model.set_data_item_computation(data_item1, computation)
            # verify setup
            self.assertEqual(data_item2, computation.get_input("a"))
            document_model.recompute_all()
            document_controller.periodic()
            # change variable
            properties = {"variable_type": "data_item", "specified_object": data_item3}
            command = Inspector.ChangeComputationVariableCommand(document_controller.document_model, computation, variable, **properties)
            command.perform()
            document_controller.push_undo_command(command)
            # verify change and trigger error
            self.assertEqual(data_item3, computation.get_input("a"))
            document_model.recompute_all()
            document_controller.periodic()
            self.assertIsNotNone(computation.error_text)
            # undo and verify
            document_controller.handle_undo()
            self.assertEqual(data_item2, computation.get_input("a"))
            document_model.recompute_all()
            document_controller.periodic()
            self.assertIsNone(computation.error_text)
            # redo and verify
            document_controller.handle_redo()
            document_model.recompute_all()
            document_controller.periodic()
            self.assertEqual(data_item3, computation.get_input("a"))
            self.assertIsNotNone(computation.error_text)

    def test_change_variable_command_resulting_in_creating_data_item_undo_redo(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            # setup
            data_item1 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item2)
            data_item3 = DataItem.DataItem(numpy.zeros((5, )))
            document_model.append_data_item(data_item3)
            computation = document_model.create_computation("target.xdata = a.xdata[6] + x")
            variable = computation.create_input_item("a", Symbolic.make_item(data_item3))
            computation.create_variable("x", value_type="integral", value=5)
            document_model.set_data_item_computation(data_item1, computation)
            # verify setup
            self.assertEqual(data_item3, computation.get_input("a"))
            document_model.recompute_all()
            document_controller.periodic()
            self.assertIsNotNone(computation.error_text)
            # change variable
            properties = {"variable_type": "data_item", "specified_object": data_item2}
            command = Inspector.ChangeComputationVariableCommand(document_controller.document_model, computation, variable, **properties)
            command.perform()
            document_controller.push_undo_command(command)
            # verify change and trigger computation
            self.assertEqual(data_item2, computation.get_input("a"))
            document_model.recompute_all()
            document_controller.periodic()
            self.assertIsNone(computation.error_text)
            # undo and verify
            document_controller.handle_undo()
            self.assertEqual(data_item3, computation.get_input("a"))
            document_model.recompute_all()
            document_controller.periodic()
            self.assertIsNotNone(computation.error_text)
            # redo and verify
            document_controller.handle_redo()
            document_model.recompute_all()
            document_controller.periodic()
            self.assertEqual(data_item2, computation.get_input("a"))
            self.assertIsNone(computation.error_text)

    def test_computation_inspector_panel_handles_computation_being_removed(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.zeros((4, 4)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            document_model.get_gaussian_blur_new(display_item, data_item, None)
            computation = document_model.computations[0]
            with contextlib.closing(ComputationPanel.InspectComputationDialog(document_controller, computation)) as dialog:
                document_controller.periodic()
                dialog.window.periodic()
                document_model.remove_computation(computation)
                document_controller.periodic()
                dialog.window.periodic()

    def test_computation_inspector_panel_handles_computation_being_removed_implicitly(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.zeros((10, )))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)

            interval = Graphics.IntervalGraphic()
            display_item.add_graphic(interval)
            interval2 = Graphics.IntervalGraphic()
            display_item.add_graphic(interval2)

            data_item2 = DataItem.DataItem(numpy.zeros((10, )))
            document_model.append_data_item(data_item2)
            display_item2 = document_model.get_display_item_for_data_item(data_item2)
            data_item3 = DataItem.DataItem(numpy.zeros((10, )))
            document_model.append_data_item(data_item3)
            display_item3 = document_model.get_display_item_for_data_item(data_item3)

            computation = document_model.create_computation()
            computation.create_input_item("src", Symbolic.make_item(data_item))
            computation.create_input_item("interval", Symbolic.make_item(interval))
            computation.create_input_item("interval2", Symbolic.make_item(interval2))
            computation.create_output_item("dst", Symbolic.make_item(data_item2))
            computation.create_output_item("dst2", Symbolic.make_item(data_item3))
            document_model.append_computation(computation)
            interval2.source = interval
            display_item.append_display_data_channel_for_data_item(data_item2)
            display_item.append_display_data_channel_for_data_item(data_item3)

            with contextlib.closing(ComputationPanel.InspectComputationDialog(document_controller, computation)):
                display_item.remove_graphic(interval).close()

    def test_remove_computation_undo_redo_cycle(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            fft_display_item = document_model.get_fft_new(display_item, data_item)
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(document_model.display_items))
            self.assertEqual(1, len(document_model.computations))
            # remove computation
            command = ComputationPanel.RemoveComputationCommand(document_controller, document_model.computations[0])
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(1, len(document_model.data_items))
            self.assertEqual(1, len(document_model.display_items))
            self.assertEqual(0, len(document_model.computations))
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(document_model.display_items))
            self.assertEqual(1, len(document_model.computations))
            # redo and check
            document_controller.handle_redo()
            self.assertEqual(1, len(document_model.data_items))
            self.assertEqual(1, len(document_model.display_items))
            self.assertEqual(0, len(document_model.computations))

    def disabled_test_expression_updates_when_variable_is_assigned(self):
        raise Exception()

    def disabled_test_computation_panel_provides_help_button(self):
        raise Exception()

    def disabled_test_new_button_create_new_data_item(self):
        raise Exception()

    def disabled_test_invalid_expression_saves_text_for_editing(self):
        raise Exception()

    def disabled_test_knobs_for_computations_appear_in_inspector(self):
        assert False



# standard libraries
import logging
import math
import typing
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift import Application
from nion.swift.model import DataItem
from nion.swift.model import Graphics
from nion.swift.test import TestContext
from nion.ui import CanvasItem
from nion.ui import DrawingContext
from nion.ui import TestUI
from nion.utils import Geometry




class TestImageCanvasItemClass(unittest.TestCase):

    def setUp(self):
        TestContext.begin_leaks()
        self._test_setup = TestContext.TestSetup()

    def tearDown(self):
        self._test_setup = typing.cast(typing.Any, None)
        TestContext.end_leaks(self)

    # Wrapper function allowing comparison of Tuples with the AlmostEqual unit test function
    def assertTupleAlmostEqual(self, tuple1: typing.Tuple[typing.Any, ...], tuple2: typing.Tuple[typing.Any, ...],
                               delta: typing.Any=None, places: typing.Any=None):
        self.assertEqual(len(tuple1), len(tuple2), "Tuples are of different lengths.")
        for a, b in zip(tuple1, tuple2):
            if delta is not None:
                self.assertAlmostEqual(a, b, delta=delta)
            elif places is not None:
                self.assertAlmostEqual(a, b, places=places)
            else:
                self.assertAlmostEqual(a, b)

    def test_mapping_widget_to_image_on_2d_data_stack_uses_signal_dimensions(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            document_controller.selected_display_panel.change_display_panel_content({"type": "image"})
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.ones((50, 10, 10)), data_descriptor=DataAndMetadata.DataDescriptor(True, 0, 2))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((100 + header_height, 100))
            # run test
            document_controller.tool_mode = "line-profile"
            display_panel.display_canvas_item.simulate_drag((20,25), (65,85))
            document_controller.periodic()
            self.assertEqual(display_item.graphics[0].vector, ((0.2, 0.25), (0.65, 0.85)))

    def test_mapping_widget_to_image_on_3d_spectrum_image_uses_collection_dimensions(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            document_controller.selected_display_panel.change_display_panel_content({"type": "image"})
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.ones((10, 10, 50)), data_descriptor=DataAndMetadata.DataDescriptor(False, 2, 1))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((100 + header_height, 100))
            # run test
            document_controller.tool_mode = "line-profile"
            display_panel.display_canvas_item.simulate_drag((20,25), (65,85))
            document_controller.periodic()
            self.assertEqual(display_item.graphics[0].vector, ((0.2, 0.25), (0.65, 0.85)))

    def test_dimension_used_for_scale_marker_on_2d_data_stack_is_correct(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            calibrations = [Calibration.Calibration(units="s"), Calibration.Calibration(units="y"), Calibration.Calibration(units="x")]
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.ones((50, 10, 10)), dimensional_calibrations=calibrations, data_descriptor=DataAndMetadata.DataDescriptor(True, 0, 2))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            self.assertEqual(display_panel.display_canvas_item._scale_marker_canvas_item_for_test._dimension_calibration_for_test.units, "x")

    def test_dimension_used_for_scale_marker_on_3d_spectrum_image_is_correct(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            calibrations = [Calibration.Calibration(units="y"), Calibration.Calibration(units="x"), Calibration.Calibration(units="e")]
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.ones((10, 10, 50)), dimensional_calibrations=calibrations, data_descriptor=DataAndMetadata.DataDescriptor(False, 2, 1))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            self.assertEqual(display_panel.display_canvas_item._scale_marker_canvas_item_for_test._dimension_calibration_for_test.units, "x")

    def test_dimension_used_for_scale_marker_on_4d_diffraction_image_is_correct(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            calibrations = [Calibration.Calibration(units="y"), Calibration.Calibration(units="x"), Calibration.Calibration(units="a"), Calibration.Calibration(units="b")]
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.ones((10, 10, 50, 50)), dimensional_calibrations=calibrations, data_descriptor=DataAndMetadata.DataDescriptor(False, 2, 2))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            self.assertEqual(display_panel.display_canvas_item._scale_marker_canvas_item_for_test._dimension_calibration_for_test.units, "b")

    def test_scale_marker_with_invalid_calibration(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            calibrations = [Calibration.Calibration(scale=math.nan), Calibration.Calibration(scale=math.nan)]
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.ones((10, 10)), dimensional_calibrations=calibrations)
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))

    def test_tool_returns_to_pointer_after_but_not_during_creating_rectangle(self):
        # setup
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            self.assertEqual(document_controller.tool_mode, "pointer")
            for tool_mode in ["rectangle", "point", "ellipse", "line"]:
                document_controller.tool_mode = tool_mode
                display_panel.display_canvas_item.simulate_press((100,125))
                display_panel.display_canvas_item.simulate_move((100,125))
                self.assertEqual(document_controller.tool_mode, tool_mode)
                display_panel.display_canvas_item.simulate_move((250,200))
                display_panel.display_canvas_item.simulate_release((250,200))
                self.assertEqual(document_controller.tool_mode, "pointer")

    def test_selected_item_takes_priority_over_all_part(self):
        # setup
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            rect_region = Graphics.RectangleGraphic()
            rect_region.bounds = (0.25, 0.25), (0.5, 0.5)
            line_region = Graphics.LineGraphic()
            line_region.start = (0.0, 1.0)
            line_region.end = (0.75, 0.25)
            # draws line, then rect
            display_item.add_graphic(line_region)
            display_item.add_graphic(rect_region)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.display_canvas_item.simulate_click((50, 950))
            document_controller.periodic()
            self.assertEqual(display_item.graphic_selection.indexes, set((0, )))
            display_panel.display_canvas_item.simulate_click((500, 500))
            document_controller.periodic()
            self.assertEqual(display_item.graphic_selection.indexes, set((0, )))

    def test_specific_parts_take_priority_over_all_part(self):
        # setup
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            rect_region = Graphics.RectangleGraphic()
            rect_region.bounds = (0.25, 0.25), (0.5, 0.5)
            line_region = Graphics.LineGraphic()
            line_region.start = (0.5, 0.5)
            line_region.end = (0.5, 1.0)
            # draws line, then rect
            display_item.add_graphic(line_region)
            display_item.add_graphic(rect_region)
            display_item = document_model.get_display_item_for_data_item(data_item)
            # clicking on line should select it
            display_panel.display_canvas_item.simulate_click((500, 600))
            document_controller.periodic()
            self.assertEqual(display_item.graphic_selection.indexes, set((0, )))

    def test_specific_parts_take_priority_when_another_selected(self):
        # setup
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            rect_region1 = Graphics.RectangleGraphic()
            rect_region1.bounds = (0.2, 0.2), (0.4, 0.4)
            rect_region2 = Graphics.RectangleGraphic()
            rect_region2.bounds = (0.4, 0.4), (0.4, 0.4)
            display_item.add_graphic(rect_region1)
            display_item.add_graphic(rect_region2)
            display_item = document_model.get_display_item_for_data_item(data_item)
            # clicking on line should select it
            display_panel.display_canvas_item.simulate_click((700, 700))
            document_controller.periodic()
            self.assertEqual(display_item.graphic_selection.indexes, set((1, )))
            display_panel.display_canvas_item.simulate_click((600, 200))
            document_controller.periodic()
            self.assertEqual(display_item.graphic_selection.indexes, set((0, )))

    def test_hit_testing_occurs_same_as_draw_order(self):
        # draw order occurs from 0 -> n
        # setup
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            rect_region1 = Graphics.RectangleGraphic()
            rect_region1.bounds = (0.2, 0.2), (0.4, 0.4)
            rect_region2 = Graphics.RectangleGraphic()
            rect_region2.bounds = (0.4, 0.4), (0.4, 0.4)
            display_item.add_graphic(rect_region1)
            display_item.add_graphic(rect_region2)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.display_canvas_item.simulate_click((500, 500))
            document_controller.periodic()
            self.assertEqual(display_item.graphic_selection.indexes, set((1, )))

    def test_1d_data_displayed_as_2d(self):
        # setup
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, )))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            display_item.display_type = "image"
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            drawing_context = DrawingContext.DrawingContext()
            display_panel.root_container.repaint_immediate(drawing_context, display_panel.root_container.canvas_size)

    def test_hand_tool_on_one_image_of_multiple_displays(self):
        # setup
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            copy_display_item = document_model.get_display_item_copy_new(display_item)
            display_panel.set_display_panel_display_item(copy_display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            # run test
            document_controller.tool_mode = "hand"
            display_panel.display_canvas_item.simulate_press((100,125))

    def test_zoom_tool_in_and_out_around_clicked_point_fit_mode(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((50, 50)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((100 + header_height, 100))
            display_panel.perform_action("set_fit_mode")
            # run test. each data pixel will be spanned by two display pixels. clicking at an odd value
            # will be a click in the center of a data pixel. zoom should expand and contract around that
            # pixel. there is also a check to make sure zooming is not just ending up with the same exact
            # zoom, which would also succeed.
            self.assertEqual((25, 25),
                             display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(51, 51)))
            self.assertEqual((30, 30),
                             display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(61, 61)))
            document_controller.tool_mode = "zoom"
            display_panel.display_canvas_item.simulate_click((20, 20))  # 10,10 in image space
            # the zoom tool runs asynchronously, so give it a slice of async time.
            document_controller.periodic()
            # check results
            # After zooming in on 20,20 canvas (10,10 image) the canvas will have zoomed in 1.25X so the display
            # now only shows 40x40 of the image on the canvas. Since it was centered around the 10,10 position of
            # the image the Canvas is now displaying 2->42 of the image with 2 image pixels off the side to the left,
            # and 8 to the right.  A click at 51,51 (image centre) now is 51/2.5 (20) image pixels in from the left,
            # plus the 2 extra based on which bit of the image we are viewing, so 20+2 = 22.

            self.assertTupleAlmostEqual((22, 22),
                                        display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(51, 51)),
                                        delta=1)
            # A click at 61,61 should NOT be that 22 plus the 5 from before. This second test verifies that we are
            # zoomed in and not just translated
            self.assertNotEqual((30, 30),
                                display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(61, 61)))

            # zoom out is done by passing the alt key to the canvas item.
            display_panel.display_canvas_item.simulate_click((20, 20), CanvasItem.KeyboardModifiers(alt=True))
            document_controller.periodic()
            # zoom in at the center of a data pixel followed by zoom out in the same spot should end up in the same original mapping.
            self.assertEqual((25, 25),
                             display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(51, 51)))

            self.assertEqual((30, 30),
                             display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(61, 61)))

    def test_zoom_tool_in_and_out_around_clicked_point_1_to_1_mode(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((40, 40)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((100 + header_height, 100))
            display_panel.perform_action("set_one_to_one_mode")
            # run test. Each canvas pixel starts as 1 image pixel, with image centered
            self.assertEqual((20, 20),
                             display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(50, 50)))
            self.assertEqual((30, 30),
                             display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(60, 60)))
            document_controller.tool_mode = "zoom"
            display_panel.display_canvas_item.simulate_click((40, 40))  # 10,10 in image space
            # the zoom tool runs asynchronously, so give it a slice of async time.
            document_controller.periodic()
            # check results
            # After zooming in on 40,40 canvas (10,10 image) the canvas will have zoomed in 1.25X so the display
            # displays the full new 50x50, but a quarter of the new pixels (2.5) are on the left and 3/4 on the right.
            # So the canvas starts displaying at 27.5 to 77.5 the image which is now 50 canvas pixels wide
            # Clicking centrally again, pixel 50 is 45% of the image across. 45% of 40 pixels is 18

            self.assertTupleAlmostEqual((18, 18),
                                        display_panel.display_canvas_item.map_widget_to_image(
                                            Geometry.IntPoint(50, 50)),
                                        delta=1)
            # A click at 61,61 should NOT be that 22 plus the 5 from before. This second test verifies that we are
            # zoomed in and not just translated
            self.assertNotEqual((35, 35),
                                display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(60, 60)))

            # zoom out is done by passing the alt key to the canvas item.
            display_panel.display_canvas_item.simulate_click((40, 40), CanvasItem.KeyboardModifiers(alt=True))
            document_controller.periodic()
            # zoom in at the center of a data pixel followed by zoom out in the same spot should end up in the
            # same original mapping.
            self.assertTupleAlmostEqual((20, 20),
                                        display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(50, 50)),
                                        delta=1)

            self.assertTupleAlmostEqual((30, 30),
                                        display_panel.display_canvas_item.map_widget_to_image(Geometry.IntPoint(60, 60)),
                                        delta=1)


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



from __future__ import annotations

# standard libraries
import abc
import collections
import copy
import functools
import itertools
import operator
import typing

# third party libraries
# None

# local libraries
from nion.utils import Event
from nion.utils import ListModel
from nion.utils import Observable


ItemValue = typing.Any


class AbstractItemSource(abc.ABC):

    @property
    @abc.abstractmethod
    def item_changed_event(self) -> Event.Event: ...

    @abc.abstractmethod
    def close(self) -> None: ...

    @property
    @abc.abstractmethod
    def item(self) -> ItemValue: ...


class AbstractItemSequenceSource(abc.ABC):

    @property
    @abc.abstractmethod
    def item_inserted_event(self) -> Event.Event: ...

    @property
    @abc.abstractmethod
    def item_removed_event(self) -> Event.Event: ...

    @property
    @abc.abstractmethod
    def item_mutated_event(self) -> Event.Event: ...

    @abc.abstractmethod
    def close(self) -> None: ...

    @property
    @abc.abstractmethod
    def items(self) -> typing.Sequence[ItemValue]: ...


class ItemSource(AbstractItemSource):
    """Provide a hard coded item source."""

    def __init__(self, item: ItemValue = None) -> None:
        self.__item = item
        self.__item_changed_event = Event.Event()

    def close(self) -> None:
        self.__item = None

    @property
    def item_changed_event(self) -> Event.Event:
        return self.__item_changed_event

    @property
    def item(self) -> ItemValue:
        return self.__item

    @item.setter
    def item(self, item: ItemValue) -> None:
        self.__item = item
        self.__item_changed_event.fire(item)


class PropertyItemSource(AbstractItemSource):
    """Provide an item source from a property of the item."""

    def __init__(self, item_source: AbstractItemSource, property: str) -> None:
        self.__item_source = item_source
        self.__property = property
        self.__last_value: ItemValue = None
        self.__item_changed_event = Event.Event()
        self.__item_source_changed_listener: typing.Optional[Event.EventListener] = None
        self.__property_changed_listener: typing.Optional[Event.EventListener] = None

        def send_value(value: typing.Optional[ItemValue]) -> None:
            if value != self.__last_value:
                self.__last_value = value
                self.__item_changed_event.fire(value)

        def item_changed(source: Observable.Observable) -> None:
            self.__unlisten()
            if source:
                def property_changed(property: str) -> None:
                    if property == self.__property:
                        send_value(getattr(source, self.__property, None))

                self.__property_changed_listener = source.property_changed_event.listen(property_changed)
                send_value(getattr(source, self.__property, None))
            else:
                send_value(None)

        self.__item_source_changed_listener = self.__item_source.item_changed_event.listen(item_changed)

        item_changed(self.__item_source.item)

    def close(self) -> None:
        self.__unlisten()
        if self.__item_source_changed_listener:
            self.__item_source_changed_listener.close()
            self.__item_source_changed_listener = None
        self.__item_source.close()
        self.__item_source = typing.cast(typing.Any, None)

    def __unlisten(self) -> None:
        if self.__property_changed_listener:
            self.__property_changed_listener.close()
            self.__property_changed_listener = None

    @property
    def item_changed_event(self) -> Event.Event:
        return self.__item_changed_event

    @property
    def item(self) -> ItemValue:
        return self.__last_value


class ArrayItemSource(AbstractItemSource):
    """Provide an item source from an array of the item."""

    def __init__(self, item_source: AbstractItemSource, property: str):
        self.__item_source = item_source
        self.__values: typing.List[ItemValue] = list()
        self.__item_changed_event = Event.Event()
        self.__item_inserted_listener: typing.Optional[Event.EventListener] = None
        self.__item_removed_listener: typing.Optional[Event.EventListener] = None

        def item_changed(source: Observable.Observable) -> None:
            self.__unlisten()
            if source:
                def item_inserted(key: str, value: ItemValue, index: int) -> None:
                    if key == property:
                        self.__values.insert(index, value)
                        self.__item_changed_event.fire(self.item)

                def item_removed(key: str, value: ItemValue, index: int) -> None:
                    if key == property:
                        self.__values.pop(index)
                        self.__item_changed_event.fire(self.item)

                self.__item_inserted_listener = source.item_inserted_event.listen(item_inserted)
                self.__item_removed_listener = source.item_removed_event.listen(item_removed)

                for index, value in enumerate(getattr(source, property)):
                    item_inserted(property, value, index)

            self.__item_changed_event.fire(self.item)

        self.__item_source_changed_listener = self.__item_source.item_changed_event.listen(item_changed)

        item_changed(self.__item_source.item)

    def close(self) -> None:
        self.__unlisten()
        self.__item_source_changed_listener.close()
        self.__item_source_changed_listener = typing.cast(typing.Any, None)
        self.__item_source.close()
        self.__item_source = typing.cast(typing.Any, None)

    def __unlisten(self) -> None:
        if self.__item_inserted_listener:
            self.__item_inserted_listener.close()
            self.__item_inserted_listener = None
        if self.__item_removed_listener:
            self.__item_removed_listener.close()
            self.__item_removed_listener = None

    @property
    def item_changed_event(self) -> Event.Event:
        return self.__item_changed_event

    @property
    def item(self) -> ItemValue:
        return copy.copy(self.__values)


AdapterInsertedFn = typing.Callable[[typing.Any, int], None]
AdapterRemovedFn = typing.Callable[[typing.Any, int], None]

class AbstractArraySourceAdapter(abc.ABC):

    @abc.abstractmethod
    def close(self) -> None: ...

    @property
    @abc.abstractmethod
    def items(self) -> typing.Sequence[ItemValue]: ...


AdapterFactory = typing.Callable[[typing.Any, AdapterInsertedFn, AdapterRemovedFn], AbstractArraySourceAdapter]


class ArrayArraySourceAdapter(AbstractArraySourceAdapter):

    @classmethod
    def factory(cls, key: str, *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None,
                mapping: typing.Optional[typing.Callable[[ItemValue], ItemValue]] = None) -> AdapterFactory:
        def make_adapter(s: Observable.Observable, insert: AdapterInsertedFn, remove: AdapterRemovedFn) -> AbstractArraySourceAdapter:
            return ArrayArraySourceAdapter(s, insert, remove, key, predicate=predicate, mapping=mapping)
        return make_adapter

    def __init__(self, source: Observable.Observable, inserted: AdapterInsertedFn, removed: AdapterRemovedFn, key: str,
                 *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None,
                 mapping: typing.Optional[typing.Callable[[ItemValue], ItemValue]] = None) -> None:
        self.__key = key
        self.__predicate = predicate or (lambda x: bool(x))
        self.__mapping = mapping or (lambda x: x)
        self.__items: typing.OrderedDict[ItemValue, ItemValue] = collections.OrderedDict()
        self.__main_items = list[Observable.Observable]()
        self.__property_changed_listeners = list[Event.EventListener | None]()

        def property_changed(item: Observable.Observable, property: str) -> None:
            in_items = item in self.__items
            should_be_in_items = self.__predicate(item)
            if in_items != should_be_in_items:
                index = self.__main_items.index(item)
                item_removed(key, item, index)
                item_inserted(key, item, index)

        def item_inserted(key: str, item: ItemValue, index: int) -> None:
            if key == self.__key:
                # also observe when properties of the item change, since that may affect the truthiness of the predicate.
                self.__main_items.insert(index, item)
                self.__property_changed_listeners.insert(index, item.property_changed_event.listen(functools.partial(property_changed, item)) if hasattr(item, "property_changed_event") else None)
                if self.__predicate(item):
                    mapped_item = self.__mapping(item) if item else None
                    index = len(self.__items)
                    self.__items[item] = mapped_item
                    inserted(mapped_item, index)

        def item_removed(key: str, item: ItemValue, index: int) -> None:
            if key == self.__key:
                self.__main_items.pop(index)
                self.__property_changed_listeners.pop(index)
                if item in self.__items:
                    index = list(self.__items.keys()).index(item)
                    mapped_item = self.__items.pop(item)
                    removed(mapped_item, index)

        self.__item_inserted_listener = source.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = source.item_removed_event.listen(item_removed)
        for index, item in enumerate(getattr(source, self.__key)):
            item_inserted(self.__key, item, index)

    def close(self) -> None:
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__item_content_changed_listener = typing.cast(typing.Any, None)

    @property
    def items(self) -> typing.Sequence[ItemValue]:
        return list(self.__items.values())


class OrderedArrayArraySourceAdapter(AbstractArraySourceAdapter):

    @classmethod
    def factory(cls, key: str, *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None) -> AdapterFactory:
        def make_adapter(s: Observable.Observable, insert: AdapterInsertedFn, remove: AdapterRemovedFn) -> AbstractArraySourceAdapter:
            return OrderedArrayArraySourceAdapter(s, insert, remove, key, predicate=predicate)
        return make_adapter

    def __init__(self, source: Observable.Observable, inserted: AdapterInsertedFn, removed: AdapterRemovedFn, key: str,
                 *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None):
        self.__list_model = ListModel.FilteredListModel(container=source, items_key=key)
        self.__items: typing.List[ItemValue] = list()

        if predicate:
            self.__list_model.filter = ListModel.PredicateFilter(predicate)

        def item_inserted(key: str, item: ItemValue, index: int) -> None:
            self.__items.insert(index, item)
            inserted(item, index)

        def item_removed(key: str, item: ItemValue, index: int) -> None:
            self.__items.pop(index)
            removed(item, index)

        self.__item_inserted_listener = self.__list_model.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = self.__list_model.item_removed_event.listen(item_removed)

        for index, item in enumerate(self.__list_model.items):
            item_inserted(key, item, index)

    def close(self) -> None:
        self.__item_inserted_listener.close()
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener.close()
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__list_model.close()
        self.__list_model = typing.cast(typing.Any, None)
        self.__items = typing.cast(typing.Any, None)

    @property
    def items(self) -> typing.Sequence[ItemValue]:
        return copy.copy(self.__items)


class SetArraySourceAdapter(AbstractArraySourceAdapter):

    @classmethod
    def factory(cls, key: str, *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None,
                mapping: typing.Optional[typing.Callable[[ItemValue], ItemValue]] = None) -> AdapterFactory:
        def make_adapter(s: Observable.Observable, insert: AdapterInsertedFn, remove: AdapterRemovedFn) -> AbstractArraySourceAdapter:
            return SetArraySourceAdapter(s, insert, remove, key, predicate=predicate, mapping=mapping)
        return make_adapter

    def __init__(self, source: Observable.Observable, inserted: AdapterInsertedFn, removed: AdapterRemovedFn, key: str,
                 *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None,
                 mapping: typing.Optional[typing.Callable[[ItemValue], ItemValue]] = None) -> None:
        self.__key = key
        self.__predicate = predicate or (lambda x: bool(x))
        self.__mapping = mapping or (lambda x: x)
        self.__items: typing.OrderedDict[ItemValue, ItemValue] = collections.OrderedDict()

        def item_added(key: str, item: ItemValue) -> None:
            if key == self.__key and self.__predicate(item):
                mapped_item = self.__mapping(item) if item else None
                index = len(self.__items)
                self.__items[item] = mapped_item
                inserted(mapped_item, index)

        def item_discarded(key: str, item: ItemValue) -> None:
            if key == self.__key and item in self.__items:
                index = list(self.__items.keys()).index(item)
                mapped_item = self.__items.pop(item)
                removed(mapped_item, index)

        self.__item_added_listener = source.item_added_event.listen(item_added)
        self.__item_discarded_listener = source.item_discarded_event.listen(item_discarded)
        for item in getattr(source, self.__key):
            item_added(self.__key, item)

    def close(self) -> None:
        self.__item_added_listener.close()
        self.__item_added_listener = typing.cast(typing.Any, None)
        self.__item_discarded_listener.close()
        self.__item_discarded_listener = typing.cast(typing.Any, None)
        self.__items = typing.cast(typing.Any, None)

    @property
    def items(self) -> typing.Sequence[ItemValue]:
        return list(self.__items.values())


class ItemSequenceSource(AbstractItemSequenceSource):

    def __init__(self, item_source: AbstractItemSource, adapter_factory: AdapterFactory) -> None:
        self.__item_source = item_source
        self.__source: typing.Optional[Observable.Observable] = None
        self.__item_inserted_event = Event.Event()
        self.__item_removed_event = Event.Event()
        self.__item_mutated_event = Event.Event()
        self.__adapter : typing.Optional[AbstractArraySourceAdapter] = None

        def item_changed(source: Observable.Observable) -> None:
            if self.__adapter:
                self.__adapter.close()
                self.__adapter = None
            if source:
                inserted = typing.cast(AdapterInsertedFn, self.__item_inserted_event.fire)
                removed = typing.cast(AdapterRemovedFn, self.__item_removed_event.fire)
                self.__adapter = adapter_factory(source, inserted, removed)

        self.__item_source_changed_listener = self.__item_source.item_changed_event.listen(item_changed)

        item_changed(self.__item_source.item)

    def close(self) -> None:
        self.__item_source_changed_listener.close()
        self.__item_source_changed_listener = typing.cast(typing.Any, None)
        self.__item_source.close()
        self.__item_source = typing.cast(typing.Any, None)
        if self.__adapter:
            self.__adapter.close()
            self.__adapter = None

    @property
    def item_inserted_event(self) -> Event.Event:
        return self.__item_inserted_event

    @property
    def item_removed_event(self) -> Event.Event:
        return self.__item_removed_event

    @property
    def item_mutated_event(self) -> Event.Event:
        return self.__item_mutated_event

    @property
    def items(self) -> typing.Sequence[ItemValue]:
        return self.__adapter.items if self.__adapter else list()


class AbstractItemTransformer(abc.ABC):

    @abc.abstractmethod
    def close(self) -> None: ...

    @property
    @abc.abstractmethod
    def item(self) -> ItemValue: ...


TransfomerMutatedFn = typing.Callable[[AbstractItemTransformer], None]
TransformerFactory = typing.Callable[[ItemValue, TransfomerMutatedFn], AbstractItemTransformer]


class PropertyItemTransformer(AbstractItemTransformer):

    @classmethod
    def factory(cls, property: str) -> TransformerFactory:
        # mutate is a callback to the transformer
        def make_transformer(item: ItemValue, mutate: TransfomerMutatedFn) -> AbstractItemTransformer:
            return cls(item, mutate, property)
        return make_transformer

    def __init__(self, item: ItemValue, mutate: TransfomerMutatedFn, property: str) -> None:
        self.__constructed_item = PropertyItemSource(ItemSource(item), property)

        def item_changed(source: Observable.Observable) -> None:
            mutate(self)

        self.__item_changed_listener = self.__constructed_item.item_changed_event.listen(item_changed)

    def close(self) -> None:
        self.__item_changed_listener.close()
        self.__item_changed_listener = typing.cast(typing.Any, None)
        self.__constructed_item.close()
        self.__constructed_item = typing.cast(typing.Any, None)

    @property
    def item(self) -> ItemValue:
        return self.__constructed_item.item


class ArrayItemTransformer(AbstractItemTransformer):

    @classmethod
    def factory(cls, property: str) -> TransformerFactory:
        # mutate is a callback to the transformer
        def make_transformer(item: ItemValue, mutate: TransfomerMutatedFn) -> AbstractItemTransformer:
            return cls(item, mutate, property)
        return make_transformer

    def __init__(self, item: ItemValue, mutate: TransfomerMutatedFn, property: str) -> None:
        self.__constructed_item = ArrayItemSource(ItemSource(item), property)

        def item_changed(source: Observable.Observable) -> None:
            mutate(self)

        self.__item_changed_listener = self.__constructed_item.item_changed_event.listen(item_changed)

    def close(self) -> None:
        self.__item_changed_listener.close()
        self.__item_changed_listener = typing.cast(typing.Any, None)
        self.__constructed_item.close()
        self.__constructed_item = typing.cast(typing.Any, None)

    @property
    def item(self) -> ItemValue:
        return self.__constructed_item.item


class FunctionItemTransformer(AbstractItemTransformer):

    @classmethod
    def factory(cls, fn: typing.Callable[[ItemValue], ItemValue]) -> TransformerFactory:
        # mutate is a callback to the transformer
        def make_transformer(item: ItemValue, mutate: TransfomerMutatedFn) -> AbstractItemTransformer:
            return cls(item, mutate, fn)
        return make_transformer

    def __init__(self, item: ItemValue, mutate: TransfomerMutatedFn, fn: typing.Callable[[ItemValue], ItemValue]):
        self.__value = fn(item)

    def close(self) -> None:
        self.__value = None

    @property
    def item(self) -> ItemValue:
        return self.__value


class TransformedItemSource(AbstractItemSource):

    def __init__(self, item_source: AbstractItemSource, transformer_factory: TransformerFactory) -> None:
        self.__item_source = item_source
        self.__item_changed_event = Event.Event()
        self.__item_value = None
        self.__transformer = None

        def transformer_item_changed(transformer: AbstractItemTransformer) -> None:
            self.__item_changed_event.fire(transformer.item)

        def item_changed(source: Observable.Observable) -> None:
            if self.__transformer:
                self.__transformer.close()
            self.__transformer = transformer_factory(self.__item_source.item, transformer_item_changed)
            self.__item_changed_event.fire(self.__transformer.item)

        self.__item_source_changed_listener = self.__item_source.item_changed_event.listen(item_changed)

        item_changed(self.__item_source.item)

    def close(self) -> None:
        if self.__transformer:
            self.__transformer.close()
            self.__transformer = None
        self.__item_source_changed_listener.close()
        self.__item_source_changed_listener = typing.cast(typing.Any, None)
        self.__item_source.close()
        self.__item_source = typing.cast(typing.Any, None)
        self.__item_value = None

    @property
    def item_changed_event(self) -> Event.Event:
        return self.__item_changed_event

    @property
    def item(self) -> ItemValue:
        return self.__transformer.item if self.__transformer else None


class TransformedItemSequence(AbstractItemSequenceSource):

    def __init__(self, item_sequence_source: AbstractItemSequenceSource, transformer_factory: TransformerFactory) -> None:
        self.__item_sequence_source = item_sequence_source
        self.__item_inserted_event = Event.Event()
        self.__item_removed_event = Event.Event()
        self.__item_mutated_event = Event.Event()
        self.__transformers: typing.List[AbstractItemTransformer] = list()

        def item_changed(transformer: AbstractItemTransformer) -> None:
            index = self.__transformers.index(transformer)
            self.__item_mutated_event.fire(transformer.item, index)

        def item_inserted(item: ItemValue, index: int) -> None:
            transformer = transformer_factory(item, item_changed)
            self.__transformers.insert(index, transformer)
            self.__item_inserted_event.fire(transformer.item, index)

        def item_removed(item: ItemValue, index: int) -> None:
            constructed_item = self.__transformers.pop(index)
            self.__item_removed_event.fire(constructed_item.item, index)
            constructed_item.close()

        def item_mutated(item: ItemValue, index: int) -> None:
            self.__transformers[index].close()
            transformer = transformer_factory(item, item_changed)
            self.__transformers[index] = transformer
            self.__item_mutated_event.fire(transformer.item, index)

        self.__item_inserted_listener = self.__item_sequence_source.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = self.__item_sequence_source.item_removed_event.listen(item_removed)
        self.__item_mutated_listener = self.__item_sequence_source.item_mutated_event.listen(item_mutated)

        for index, item in enumerate(self.__item_sequence_source.items):
            item_inserted(item, index)

    def close(self) -> None:
        self.__item_inserted_listener.close()
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener.close()
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__item_mutated_listener.close()
        self.__item_mutated_listener = typing.cast(typing.Any, None)
        for item in self.__transformers:
            item.close()
        self.__transformers = typing.cast(typing.Any, None)
        self.__item_sequence_source.close()
        self.__item_sequence_source = typing.cast(typing.Any, None)

    @property
    def item_inserted_event(self) -> Event.Event:
        return self.__item_inserted_event

    @property
    def item_removed_event(self) -> Event.Event:
        return self.__item_removed_event

    @property
    def item_mutated_event(self) -> Event.Event:
        return self.__item_mutated_event

    @property
    def items(self) -> typing.Sequence[ItemValue]:
        return [item.item for item in self.__transformers]


class FilteredItemSequence(AbstractItemSequenceSource):

    def __init__(self, item_sequence_source: AbstractItemSequenceSource, predicate: typing.Callable[[ItemValue], bool]) -> None:
        self.__item_sequence_source = item_sequence_source
        self.__list_model = ListModel.ListModel[ItemValue](key="items")
        self.__filtered_list_model = ListModel.FilteredListModel(container=self.__list_model, items_key="items")
        self.__filtered_list_model.filter = ListModel.PredicateFilter(lambda x: predicate(x.value))
        self.__item_inserted_event = Event.Event()
        self.__item_removed_event = Event.Event()
        self.__item_mutated_event = Event.Event()

        class Wrapper:  # used to wrap values so that they are all unique
            def __init__(self, value: ItemValue) -> None: self.value = value

        def item_inserted(item: ItemValue, index: int) -> None:
            self.__list_model.insert_item(index, Wrapper(item))

        def item_removed(item: ItemValue, index: int) -> None:
            self.__list_model.remove_item(index)

        def item_mutated(item: ItemValue, index: int) -> None:
            self.__list_model.remove_item(index)
            self.__list_model.insert_item(index, Wrapper(item))

        self.__item_inserted_listener = self.__item_sequence_source.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = self.__item_sequence_source.item_removed_event.listen(item_removed)
        self.__item_mutated_listener = self.__item_sequence_source.item_mutated_event.listen(item_mutated)

        def filtered_item_inserted(key: str, item: ItemValue, index: int) -> None:
            self.__item_inserted_event.fire(typing.cast(Wrapper, item).value, index)

        def filtered_item_removed(key: str, item: ItemValue, index: int) -> None:
            self.__item_removed_event.fire(typing.cast(Wrapper, item).value, index)

        self.__filtered_item_inserted_listener = self.__filtered_list_model.item_inserted_event.listen(filtered_item_inserted)
        self.__filtered_item_removed_listener = self.__filtered_list_model.item_removed_event.listen(filtered_item_removed)

        for index, item in enumerate(self.__item_sequence_source.items):
            item_inserted(item, index)

    def close(self) -> None:
        self.__item_inserted_listener.close()
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener.close()
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__item_mutated_listener.close()
        self.__item_mutated_listener = typing.cast(typing.Any, None)
        self.__filtered_item_inserted_listener.close()
        self.__filtered_item_inserted_listener = typing.cast(typing.Any, None)
        self.__filtered_item_removed_listener.close()
        self.__filtered_item_removed_listener = typing.cast(typing.Any, None)
        self.__filtered_list_model.close()
        self.__filtered_list_model = typing.cast(typing.Any, None)
        self.__list_model.close()
        self.__list_model = typing.cast(typing.Any, None)
        self.__item_sequence_source.close()
        self.__item_sequence_source = typing.cast(typing.Any, None)

    @property
    def item_inserted_event(self) -> Event.Event:
        return self.__item_inserted_event

    @property
    def item_removed_event(self) -> Event.Event:
        return self.__item_removed_event

    @property
    def item_mutated_event(self) -> Event.Event:
        return self.__item_mutated_event

    @property
    def items(self) -> typing.Sequence[ItemValue]:
        return [item.value for item in self.__filtered_list_model.items]


class MappedItemSequence(AbstractItemSequenceSource):

    def __init__(self, item_sequence_source: AbstractItemSequenceSource, map_fn: typing.Callable[[typing.Optional[AbstractItemSource]], typing.Optional[AbstractItemSource]]):
        self.__item_sequence_source = item_sequence_source
        self.__item_inserted_event = Event.Event()
        self.__item_removed_event = Event.Event()
        self.__item_mutated_event = Event.Event()
        self.__mapped_items: typing.List[AbstractItemSource] = list()
        self.__mapped_item_listeners: typing.List[Event.EventListener] = list()

        def item_changed(mapped_item: AbstractItemSource, item: ItemValue) -> None:
            index = self.__mapped_items.index(mapped_item)
            self.__item_mutated_event.fire(mapped_item.item, index)

        def item_inserted(item: ItemValue, index: int) -> None:
            mapped_item = map_fn(ItemSource(item))
            assert mapped_item is not None
            self.__mapped_items.insert(index, mapped_item)
            self.__mapped_item_listeners.insert(index, mapped_item.item_changed_event.listen(functools.partial(item_changed, mapped_item)))
            self.__item_inserted_event.fire(mapped_item.item, index)

        def item_removed(item: ItemValue, index: int) -> None:
            self.__mapped_item_listeners.pop(index).close()
            mapped_item = self.__mapped_items.pop(index)
            self.__item_removed_event.fire(mapped_item.item, index)
            mapped_item.close()

        def item_mutated(item: ItemValue, index: int) -> None:
            self.__mapped_item_listeners[index].close()
            self.__mapped_items[index].close()
            mapped_item = map_fn(ItemSource(item))
            assert mapped_item is not None
            self.__mapped_items[index] = mapped_item
            self.__mapped_item_listeners[index] = mapped_item.item_changed_event.listen(functools.partial(item_changed, mapped_item))
            self.__item_mutated_event.fire(mapped_item.item, index)

        self.__item_inserted_listener = self.__item_sequence_source.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = self.__item_sequence_source.item_removed_event.listen(item_removed)
        self.__item_mutated_listener = self.__item_sequence_source.item_mutated_event.listen(item_mutated)

        for index, item in enumerate(self.__item_sequence_source.items):
            item_inserted(item, index)

    def close(self) -> None:
        self.__item_inserted_listener.close()
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener.close()
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__item_mutated_listener.close()
        self.__item_mutated_listener = typing.cast(typing.Any, None)
        for mapped_item_listener in self.__mapped_item_listeners:
            mapped_item_listener.close()
        self.__mapped_item_listeners = typing.cast(typing.Any, None)
        for item in self.__mapped_items:
            item.close()
        self.__mapped_items = typing.cast(typing.Any, None)
        self.__item_sequence_source.close()
        self.__item_sequence_source = typing.cast(typing.Any, None)

    @property
    def item_inserted_event(self) -> Event.Event:
        return self.__item_inserted_event

    @property
    def item_removed_event(self) -> Event.Event:
        return self.__item_removed_event

    @property
    def item_mutated_event(self) -> Event.Event:
        return self.__item_mutated_event

    @property
    def items(self) -> typing.Sequence[ItemValue]:
        return [item.item for item in self.__mapped_items]


class AbstractAction(abc.ABC):

    @abc.abstractmethod
    def close(self) -> None: ...


class ItemSequenceAction:
    """Monitor an item set source and create an action on each item."""

    def __init__(self, item_sequence_source: AbstractItemSequenceSource, action: typing.Callable[[AbstractItemSource], AbstractAction]):
        self.__item_sequence_source = item_sequence_source
        self.__item_actions: typing.Dict[ItemValue, AbstractAction] = dict()

        def item_inserted(item: ItemValue, index: int) -> None:
            self.__item_actions[item] = action(ItemSource(item))

        def item_removed(item: ItemValue, index: int) -> None:
            self.__item_actions.pop(item).close()

        def item_mutated(item: ItemValue, index: int) -> None:
            self.__item_actions[item].close()
            self.__item_actions[item] = action(ItemSource(item))

        self.__item_inserted_listener = self.__item_sequence_source.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = self.__item_sequence_source.item_removed_event.listen(item_removed)
        self.__item_mutated_listener = self.__item_sequence_source.item_mutated_event.listen(item_mutated)

        for index, item in enumerate(self.__item_sequence_source.items):
            item_inserted(item, index)

    def close(self) -> None:
        self.__item_inserted_listener.close()
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener.close()
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__item_mutated_listener.close()
        self.__item_mutated_listener = typing.cast(typing.Any, None)
        for item, item_action in self.__item_actions.items():
            item_action.close()
        self.__item_actions = typing.cast(typing.Any, None)
        self.__item_sequence_source.close()
        self.__item_sequence_source = typing.cast(typing.Any, None)


class ItemSequenceTrampoline:
    """Monitor an item set source and send the action to another object."""

    def __init__(self, item_sequence_source: AbstractItemSequenceSource, target: Observable.Observable, key: str) -> None:
        self.__item_sequence_source = item_sequence_source
        self.__items: typing.List[ItemValue] = list()

        def item_inserted(item: ItemValue, index: int) -> None:
            self.__items.insert(index, item)
            target.notify_insert_item(key, item, index)

        def item_removed(item: ItemValue, index: int) -> None:
            self.__items.pop(index)
            target.notify_remove_item(key, item, index)

        self.__item_inserted_listener = self.__item_sequence_source.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = self.__item_sequence_source.item_removed_event.listen(item_removed)

        for index, item in enumerate(self.__item_sequence_source.items):
            item_inserted(item, index)

    def close(self) -> None:
        self.__item_inserted_listener.close()
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener.close()
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__items = typing.cast(typing.Any, None)
        self.__item_sequence_source.close()
        self.__item_sequence_source = typing.cast(typing.Any, None)

    @property
    def item(self) -> ItemValue:
        return self.__items


class AbstractItemSequenceReducer(abc.ABC):

    @abc.abstractmethod
    def close(self) -> None: ...

    @abc.abstractmethod
    def item_inserted(self, item: ItemValue, index: int) -> bool: ...

    @abc.abstractmethod
    def item_removed(self, item: ItemValue, index: int) -> bool: ...

    @abc.abstractmethod
    def item_mutated(self, item: ItemValue, index: int) -> bool: ...

    @property
    @abc.abstractmethod
    def item(self) -> ItemValue: ...


class ItemSequenceCollector(AbstractItemSequenceReducer):

    def __init__(self) -> None:
        self.__items: typing.List[ItemValue] = list()

    def close(self) -> None:
        self.__items = typing.cast(typing.Any, None)

    def item_inserted(self, item: ItemValue, index: int) -> bool:
        self.__items.insert(index, item)
        return True

    def item_removed(self, item: ItemValue, index: int) -> bool:
        self.__items.pop(index)
        return True

    def item_mutated(self, item: ItemValue, index: int) -> bool:
        self.__items[index] = item
        return True

    @property
    def item(self) -> ItemValue:
        return copy.copy(self.__items)


class ItemSequenceLength(AbstractItemSequenceReducer):

    def __init__(self) -> None:
        self.__items: typing.List[ItemValue] = list()

    def close(self) -> None:
        self.__items = typing.cast(typing.Any, None)

    def item_inserted(self, item: ItemValue, index: int) -> bool:
        self.__items.insert(index, item)
        return True

    def item_removed(self, item: ItemValue, index: int) -> bool:
        self.__items.pop(index)
        return True

    def item_mutated(self, item: ItemValue, index: int) -> bool:
        return False

    @property
    def item(self) -> ItemValue:
        return len(self.__items)


class ItemSequenceListReducer(AbstractItemSource):
    """Monitor an item set source and run a reducer on it to produce an item value."""

    def __init__(self, item_sequence_source: AbstractItemSequenceSource, reducer: AbstractItemSequenceReducer) -> None:
        self.__item_sequence_source = item_sequence_source
        self.__item_reducer = reducer
        self.__item_changed_event = Event.Event()

        def item_inserted(item: ItemValue, index: int) -> None:
            if self.__item_reducer.item_inserted(item, index):
                self.item_changed_event.fire(self.item)

        def item_removed(item: ItemValue, index: int) -> None:
            if self.__item_reducer.item_removed(item, index):
                self.item_changed_event.fire(self.item)

        def item_mutated(item: ItemValue, index: int) -> None:
            if self.__item_reducer.item_mutated(item, index):
                self.item_changed_event.fire(self.item)

        self.__item_inserted_listener = self.__item_sequence_source.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = self.__item_sequence_source.item_removed_event.listen(item_removed)
        self.__item_mutated_listener = self.__item_sequence_source.item_mutated_event.listen(item_mutated)

        for index, item in enumerate(self.__item_sequence_source.items):
            item_inserted(item, index)

    def close(self) -> None:
        self.__item_inserted_listener.close()
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener.close()
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__item_mutated_listener.close()
        self.__item_mutated_listener = typing.cast(typing.Any, None)
        self.__item_sequence_source.close()
        self.__item_sequence_source = typing.cast(typing.Any, None)
        self.__item_reducer.close()
        self.__item_reducer = typing.cast(typing.Any, None)

    @property
    def item_changed_event(self) -> Event.Event:
        return self.__item_changed_event

    @property
    def item(self) -> ItemValue:
        return self.__item_reducer.item


class ItemSequenceIndex(AbstractItemSource):

    def __init__(self, item_sequence_source: AbstractItemSequenceSource, index: int) -> None:
        self.__item_sequence_source = item_sequence_source
        self.__index = index
        self.__item: typing.Optional[ItemValue] = None
        self.__item_changed_event = Event.Event()

        def update_item() -> None:
            items = self.__item_sequence_source.items
            item = items[index] if len(items) > index else None
            if item != self.__item:
                self.__item = item
                self.item_changed_event.fire(self.__item)

        def item_inserted(item: ItemValue, index: int) -> None:
            update_item()

        def item_removed(item: ItemValue, index: int) -> None:
            update_item()

        def item_mutated(item: ItemValue, index: int) -> None:
            update_item()

        self.__item_inserted_listener = self.__item_sequence_source.item_inserted_event.listen(item_inserted)
        self.__item_removed_listener = self.__item_sequence_source.item_removed_event.listen(item_removed)
        self.__item_mutated_listener = self.__item_sequence_source.item_mutated_event.listen(item_mutated)

        update_item()

    def close(self) -> None:
        self.__item_inserted_listener.close()
        self.__item_inserted_listener = typing.cast(typing.Any, None)
        self.__item_removed_listener.close()
        self.__item_removed_listener = typing.cast(typing.Any, None)
        self.__item_mutated_listener.close()
        self.__item_mutated_listener = typing.cast(typing.Any, None)
        self.__item_sequence_source.close()
        self.__item_sequence_source = typing.cast(typing.Any, None)
        self.__item = None

    @property
    def item_changed_event(self) -> Event.Event:
        return self.__item_changed_event

    @property
    def item(self) -> ItemValue:
        return self.__item


class ItemTuple(AbstractItemSource):

    def __init__(self, *items: typing.Optional[AbstractItemSource]):
        self.__items = items
        self.__item_changed_listeners: typing.List[typing.Optional[Event.EventListener]] = list()
        self.__values: typing.List[ItemValue] = list()
        self.__item_changed_event = Event.Event()

        def item_changed(index: int, item: ItemValue) -> None:
            if item != self.__values[index]:
                self.__values[index] = item
                self.item_changed_event.fire(self.item)

        for index, item in enumerate(self.__items):
            item_changed_listener = item.item_changed_event.listen(functools.partial(item_changed, index)) if item else None
            self.__item_changed_listeners.append(item_changed_listener)
            self.__values.append(item.item if item else None)

    def close(self) -> None:
        for item_changed_listener in self.__item_changed_listeners:
            if item_changed_listener:
                item_changed_listener.close()
        self.__item_changed_listeners = typing.cast(typing.Any, None)
        for item in self.__items:
            if item:
                item.close()
        self.__items = typing.cast(typing.Any, None)
        self.__values = typing.cast(typing.Any, None)

    @property
    def item_changed_event(self) -> Event.Event:
        return self.__item_changed_event

    @property
    def item(self) -> ItemValue:
        return tuple(self.__values)


class ItemAction(AbstractAction):

    def __init__(self, item_source: AbstractItemSource, action_factory: typing.Callable[[AbstractItemSource], AbstractAction]):
        self.__item_source = item_source
        self.__item_action = None
        self.__item = None

        def item_changed(item: ItemValue) -> None:
            if item != self.__item:
                if self.__item_action:
                    self.__item_action.close()
                self.__item_action = action_factory(ItemSource(item)) if item else None

        self.__item_changed_listener = item_source.item_changed_event.listen(item_changed)

        item_changed(self.__item_source.item)

    def close(self) -> None:
        if self.__item_action:
            self.__item_action.close()
        self.__item_action = None
        self.__item = None
        self.__item_changed_listener.close()
        self.__item_changed_listener = typing.cast(typing.Any, None)
        self.__item_source.close()
        self.__item_source = typing.cast(typing.Any, None)


class ItemMonitor(AbstractAction):

    def __init__(self, item_source: AbstractItemSource, fn: typing.Callable[[ItemValue], None]) -> None:
        self.__item_source = item_source

        def item_changed(item: ItemValue) -> None:
            fn(item)

        self.__item_changed_listener = item_source.item_changed_event.listen(item_changed)

        item_changed(self.__item_source.item)

    def close(self) -> None:
        self.__item_changed_listener.close()
        self.__item_changed_listener = typing.cast(typing.Any, None)
        self.__item_source.close()
        self.__item_source = typing.cast(typing.Any, None)


class ItemValueModel(Observable.Observable):
    """A value model based on an item source.

    This can be used directly with with Observer.ItemValueModel(oo.make_observable()) as a model.
    """

    def __init__(self, item_source: AbstractItemSource | None) -> None:
        super().__init__()

        # item_source must be valid. however, the type signature of this init specifies optional so that it can
        # be directly used with make_observable() which returns None if the item source is not valid.
        assert item_source

        self.__item_source = item_source
        self.__item = None

        def item_changed(item: ItemValue) -> None:
            if item != self.__item:
                self.__item = item
                self.property_changed_event.fire("value")

        self.__item_changed_listener = item_source.item_changed_event.listen(item_changed)

        item_changed(item_source.item)

    @property
    def value(self) -> ItemValue:
        return self.__item

    def close(self) -> None:
        self.__item_source.close()
        self.__item_source = typing.cast(typing.Any, None)


class ObserverBuilder:
    """The observer builder base. Constructs an observable. Provides root item."""

    def __init__(self) -> None:
        self.__fns: typing.List[typing.Callable[[ItemValue], AbstractItemSource]] = list()

    def close(self) -> None:
        self.__fns = typing.cast(typing.Any, None)

    def make_observable(self) -> typing.Optional[AbstractItemSource]:
        """Returns an observable, which must be closed."""
        return self._apply(None)

    def _apply(self, node: typing.Optional[AbstractItemSource]) -> typing.Optional[AbstractItemSource]:
        o = node
        for fn in self.__fns:
            o = fn(o)
        return o

    def _build(self, fn: typing.Callable[[ItemValue], AbstractItemSource]) -> None:
        self.__fns.append(fn)

    def source(self, item_value: ItemValue) -> ObserverBuilderItemSource:
        """Build an observer source from the item_value."""
        self._build(lambda node: ItemSource(item_value))
        return ObserverBuilderItemSource(self)

    @property
    def x(self) -> ObserverBuilderItemSource:
        """Build an observer item source base, used as input to functions iterating over a collection."""
        return ObserverBuilderItemSource(ObserverBuilder())


class ObserverBuilderItemSource:
    """The observer builder for item sources."""

    def __init__(self, base: ObserverBuilder):
        self.__base = base

    def _build(self, fn: typing.Callable[[typing.Any], typing.Any]) -> None:
        self.__base._build(fn)

    @property
    def base(self) -> ObserverBuilder:
        return self.__base

    def sequence_from_array(self, key: str, *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None) -> ObserverBuilderItemSequenceSource:
        """Builds an unordered item sequence from the observable array specified by key and filtered by predicate."""
        self._build(lambda node: ItemSequenceSource(typing.cast(AbstractItemSource, node), ArrayArraySourceAdapter.factory(key, predicate=predicate)))
        return ObserverBuilderItemSequenceSource(self.base)

    def ordered_sequence_from_array(self, key: str, *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None) -> ObserverBuilderItemSequenceSource:
        """Builds an ordered item sequence from the observable array specified by key and filtered by predicate. Slower."""
        self._build(lambda node: ItemSequenceSource(typing.cast(AbstractItemSource, node), OrderedArrayArraySourceAdapter.factory(key, predicate=predicate)))
        return ObserverBuilderItemSequenceSource(self.base)

    def sequence_from_set(self, key: str, *, predicate: typing.Optional[typing.Callable[[ItemValue], bool]] = None) -> ObserverBuilderItemSequenceSource:
        """Builds an unordered item sequence from the observable set specified by key and filtered by predicate."""
        self._build(lambda node: ItemSequenceSource(typing.cast(AbstractItemSource, node), SetArraySourceAdapter.factory(key, predicate=predicate)))
        return ObserverBuilderItemSequenceSource(self.base)

    def print(self) -> ObserverBuilderItemAction:
        """Build an action to print the item."""
        self._build(lambda node: ItemMonitor(typing.cast(AbstractItemSource, node), print))
        return ObserverBuilderItemAction(self.base)

    def action(self, action: typing.Callable[[ItemValue], AbstractAction]) -> ObserverBuilderItemAction:
        """Build an action on the item."""
        def make_action(node: AbstractItemSource) -> AbstractAction:
            return action(node.item)
        self._build(lambda node: ItemAction(typing.cast(AbstractItemSource, node), make_action))
        return ObserverBuilderItemAction(self.base)

    def action_fn(self, item_changed: typing.Callable[[ItemValue], None]) -> ObserverBuilderItemAction:
        """Build an action on the item."""
        self._build(lambda node: ItemMonitor(typing.cast(AbstractItemSource, node), item_changed))
        return ObserverBuilderItemAction(self.base)

    def prop(self, property: str) -> ObserverBuilderItemSource:
        """Build a property observer on the item. The item will be the property."""
        self._build(lambda node: TransformedItemSource(typing.cast(AbstractItemSource, node), PropertyItemTransformer.factory(property)))
        return ObserverBuilderItemSource(self.base)

    def array(self, property: str) -> ObserverBuilderItemSource:
        """Build a array observer on the item. The item will be an array."""
        self._build(lambda node: TransformedItemSource(typing.cast(AbstractItemSource, node), ArrayItemTransformer.factory(property)))
        return ObserverBuilderItemSource(self.base)

    def getter(self, op: typing.Callable[[ItemValue], ItemValue]) -> ObserverBuilderItemSource:
        """Build a getter on the item. The item will be the result of the getter on the value."""
        self._build(lambda node: ItemSource(op(node.item) if node.item else None))
        return ObserverBuilderItemSource(self.base)

    def get(self, attribute: str) -> ObserverBuilderItemSource:
        """Build an unobserved property value on the item. The item will be the property."""
        return self.getter(operator.attrgetter(attribute))

    def tuple(self, *items: ObserverBuilderItemSource) -> ObserverBuilderItemSource:
        """Build a tuple from a list of items. The item will be a tuple."""
        def make_tuple(node: AbstractItemSource) -> ItemTuple:
            item_sources = [item.base._apply(node) for item in items]
            return ItemTuple(*item_sources)
        self._build(make_tuple)
        return ObserverBuilderItemSource(self.base)

    def constant(self, value: ItemValue) -> ObserverBuilderItemSource:
        """Build a constant value. The item will be the constant."""
        self._build(lambda node: ItemSource(value))
        return ObserverBuilderItemSource(self.base)

    def transform(self, fn: typing.Callable[[ItemValue], ItemValue]) -> ObserverBuilderItemSource:
        """Build a transformer applied to the value. The item will be the transformed item."""
        self._build(lambda node: TransformedItemSource(node, FunctionItemTransformer.factory(fn)))
        return ObserverBuilderItemSource(self.base)

    def flatten(self) -> ObserverBuilderItemSource:
        """Build a flattened list. The item will be the list of all the elements in the list of lists."""
        return self.transform(lambda x: list(itertools.chain(*x)))


class ObserverBuilderItemSequenceSource:
    """The observer builder for item sequence sources."""

    def __init__(self, base: ObserverBuilder):
        self.__base = base

    def _build(self, fn: typing.Callable[[typing.Any], typing.Any]) -> None:
        self.__base._build(fn)

    @property
    def base(self) -> ObserverBuilder:
        return self.__base

    def for_each(self, item: ObserverBuilderItemAction) -> ObserverBuilderItemAction:
        """Perform an action on each item in the sequence."""
        def action(node: AbstractItemSource) -> AbstractAction:
            return typing.cast(AbstractAction, item.base._apply(node))
        self._build(lambda node: ItemSequenceAction(typing.cast(AbstractItemSequenceSource, node), action))
        return ObserverBuilderItemAction(self.base)

    def map(self, item: ObserverBuilderItemSource) -> ObserverBuilderItemSequenceSource:
        """Perform a mapping on each item in the sequence. The items will be a mapping of the input sequence."""
        def action(node: typing.Optional[AbstractItemSource]) -> typing.Optional[AbstractItemSource]:
            return item.base._apply(node)
        self._build(lambda node: MappedItemSequence(typing.cast(AbstractItemSequenceSource, node), action))
        return ObserverBuilderItemSequenceSource(self.base)

    def filter(self, predicate: typing.Callable[[ItemValue], bool]) -> ObserverBuilderItemSequenceSource:
        """Perform a filter on each item in the sequence. The items will be a filtered version of the input sequence."""
        self._build(lambda node: FilteredItemSequence(typing.cast(AbstractItemSequenceSource, node), predicate))
        return ObserverBuilderItemSequenceSource(self.base)

    def trampoline(self, target: Observable.Observable, key: str) -> ObserverBuilderItemAction:
        """Observe the sequence and pass along changes to the target as a sequence."""
        self._build(lambda node: ItemSequenceTrampoline(typing.cast(AbstractItemSequenceSource, node), target, key))
        return ObserverBuilderItemAction(self.base)

    def collect_list(self) -> ObserverBuilderItemSource:
        """Convert the item sequence to a list. The item will be a list of the inputs."""
        self._build(lambda node: ItemSequenceListReducer(typing.cast(AbstractItemSequenceSource, node), ItemSequenceCollector()))
        return ObserverBuilderItemSource(self.base)

    def len(self) -> ObserverBuilderItemSource:
        """Return the length of the sequence."""
        self._build(lambda node: ItemSequenceListReducer(typing.cast(AbstractItemSequenceSource, node), ItemSequenceLength()))
        return ObserverBuilderItemSource(self.base)

    def index(self, index: int) -> ObserverBuilderItemSource:
        """Return the index item in the ordered sequence."""
        self._build(lambda node: ItemSequenceIndex(typing.cast(AbstractItemSequenceSource, node), index))
        return ObserverBuilderItemSource(self.base)

    def sequence_property(self, property: str) -> ObserverBuilderItemSequenceSource:
        self._build(lambda node: TransformedItemSequence(node, PropertyItemTransformer.factory(property)))
        return ObserverBuilderItemSequenceSource(self.base)

    def sequence_array(self, key: str) -> ObserverBuilderItemSequenceSource:
        self._build(lambda node: TransformedItemSequence(node, ArrayItemTransformer.factory(key)))
        return ObserverBuilderItemSequenceSource(self.base)


class ObserverBuilderItemAction:

    def __init__(self, base: ObserverBuilder):
        self.__base = base

    def _build(self, fn: typing.Callable[[typing.Any], typing.Any]) -> None:
        self.__base._build(fn)

    @property
    def base(self) -> ObserverBuilder:
        return self.__base
